/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

/*
 * Copyright (c) 2004-2006 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2007 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#include "ompi_config.h"

#include "opal/class/opal_hash_table.h"
#include "ompi/mca/btl/base/btl_base_error.h"
#include "ompi/runtime/ompi_module_exchange.h"
#include "opal/util/arch.h"
#include "opal/util/if.h"
#include "opal/util/net.h"
#include "opal/include/opal/opal_socket_errno.h"

#include "btl_eth.h"
#include "btl_eth_proc.h"
 
#include "af_enet.h"

/* shall debugging messages be printed about the local interface lookup process */
#define BTL_ETH_DEBUG_IFLOOKUP 0

static void mca_btl_eth_proc_construct(mca_btl_eth_proc_t* proc);
static void mca_btl_eth_proc_destruct(mca_btl_eth_proc_t* proc);

mca_btl_eth_interface_t* local_interfaces[MAX_KERNEL_INTERFACES];
mca_btl_eth_interface_t* peer_interfaces[MAX_KERNEL_INTERFACES];
int local_kindex_to_index[MAX_KERNEL_INTERFACE_INDEX];
int peer_kindex_to_index[MAX_KERNEL_INTERFACE_INDEX];
size_t num_local_interfaces;
size_t num_peer_interfaces;
unsigned int *best_assignment;
int max_assignment_weight;
int max_assignment_cardinality;
enum mca_btl_eth_connection_quality **weights;
struct mca_btl_eth_addr_t ***best_addr;

OBJ_CLASS_INSTANCE( mca_btl_eth_proc_t, 
                    opal_list_item_t, 
                    mca_btl_eth_proc_construct, 
                    mca_btl_eth_proc_destruct );

void mca_btl_eth_proc_construct(mca_btl_eth_proc_t* proc)
{
    proc->proc_ompi = 0;
    proc->proc_addrs = NULL;
    proc->proc_addr_count = 0;
    proc->proc_endpoints = NULL;
    proc->proc_endpoint_count = 0;
    OBJ_CONSTRUCT(&proc->proc_lock, opal_mutex_t);
}

/*
 * Cleanup ib proc instance
 */

void mca_btl_eth_proc_destruct(mca_btl_eth_proc_t* proc)
{
    /* remove from list of all proc instances */
    OPAL_THREAD_LOCK(&mca_btl_eth_component.eth_lock);
    opal_hash_table_remove_value_uint64(&mca_btl_eth_component.eth_procs, 
                                        orte_util_hash_name(&proc->proc_name));
    OPAL_THREAD_UNLOCK(&mca_btl_eth_component.eth_lock);

    /* release resources */
    if(NULL != proc->proc_endpoints) {
        free(proc->proc_endpoints);
    }
    OBJ_DESTRUCT(&proc->proc_lock);
}

/*
 * Create a ETH process structure. There is a one-to-one correspondence
 * between a ompi_proc_t and a mca_btl_eth_proc_t instance. We cache
 * additional data (specifically the list of mca_btl_eth_endpoint_t instances, 
 * and published addresses) associated w/ a given destination on this
 * datastructure.
 */

mca_btl_eth_proc_t* mca_btl_eth_proc_create(ompi_proc_t* ompi_proc)
{
    int rc;
    size_t size;
    mca_btl_eth_proc_t* btl_proc;
    uint64_t hash = orte_util_hash_name(&ompi_proc->proc_name);

    OPAL_THREAD_LOCK(&mca_btl_eth_component.eth_lock);
    rc = opal_hash_table_get_value_uint64(&mca_btl_eth_component.eth_procs, 
                                          hash, (void**)&btl_proc);
    if(OMPI_SUCCESS == rc) {
        OPAL_THREAD_UNLOCK(&mca_btl_eth_component.eth_lock);
        return btl_proc;
    }

    btl_proc = OBJ_NEW(mca_btl_eth_proc_t);
    if(NULL == btl_proc)
        return NULL;
    btl_proc->proc_ompi = ompi_proc;
    btl_proc->proc_name = ompi_proc->proc_name;
    
    /* add to hash table of all proc instance */
    opal_hash_table_set_value_uint64(&mca_btl_eth_component.eth_procs,
                                     hash, btl_proc);
    OPAL_THREAD_UNLOCK(&mca_btl_eth_component.eth_lock);

    /* lookup eth parameters exported by this proc */
    rc = ompi_modex_recv( &mca_btl_eth_component.super.btl_version,
                                  ompi_proc,
                                  (void**)&btl_proc->proc_addrs,
                                  &size );
    if(rc != OMPI_SUCCESS) {
        BTL_ERROR(("mca_base_modex_recv: failed with return value=%d", rc));
        OBJ_RELEASE(btl_proc);
        return NULL;
    }
    if(0 != (size % sizeof(mca_btl_eth_addr_t))) {
        BTL_ERROR(("mca_base_modex_recv: invalid size %lu: btl-size: %lu\n",
          (unsigned long) size, (unsigned long) sizeof(mca_btl_eth_addr_t)));
        return NULL;
    }
    btl_proc->proc_addr_count = size / sizeof(mca_btl_eth_addr_t);

    /* allocate space for endpoint array - one for each exported address */
    btl_proc->proc_endpoints = (mca_btl_base_endpoint_t**)
        malloc((1 + btl_proc->proc_addr_count) *
                sizeof(mca_btl_base_endpoint_t*));
    if(NULL == btl_proc->proc_endpoints) {
        OBJ_RELEASE(btl_proc);
        return NULL;
    }
    if(NULL == mca_btl_eth_component.eth_local && ompi_proc == ompi_proc_local()) {
        mca_btl_eth_component.eth_local = btl_proc;
    }
    return btl_proc;
}


/*
 * Note that this routine must be called with the lock on the process
 * already held.  Insert a btl instance into the proc array and assign 
 * it an address.
 */
int mca_btl_eth_proc_insert( mca_btl_eth_proc_t* btl_proc, 
                             mca_btl_base_endpoint_t* btl_endpoint )
{
    struct mca_btl_eth_module_t *btl_eth = btl_endpoint->endpoint_btl;
    size_t i;
    int s;
    struct timeval timeout;

#ifndef WORDS_BIGENDIAN
    /* if we are little endian and our peer is not so lucky, then we
       need to put all information sent to him in big endian (aka
       Network Byte Order) and expect all information received to
       be in NBO.  Since big endian machines always send and receive
       in NBO, we don't care so much about that case. */
    if (btl_proc->proc_ompi->proc_arch & OPAL_ARCH_ISBIGENDIAN) {
        btl_endpoint->endpoint_nbo = true;
    }
#endif

    /* insert into endpoint array */
    btl_endpoint->endpoint_proc = btl_proc;
    btl_proc->proc_endpoints[btl_proc->proc_endpoint_count++] = btl_endpoint;

    /* The local proc won't be reachable via this BTL - so skip it */
    if (mca_btl_eth_component.eth_local == btl_proc) {
#if BTL_ETH_DEBUG_IFLOOKUP
    	BTL_OUTPUT(("skipped checking local proc (%d %d) via if %d",
            btl_proc->proc_name.jobid,
            btl_proc->proc_name.vpid,
            btl_eth->eth_ifkindex));
#endif
        return OMPI_ERR_UNREACH;
    }

    /*
     * We already know the Ethernet address(es) and port(s) of the remote
     * process, but we still have to figure out whether it can be reached via
     * the interface this btl module has bound to.
     */
    s = socket(AF_ENET, SOCK_DGRAM, EPPROTO_EDP);
    if (s <= 0) {
        BTL_ERROR(("socket(): failed: %s (%d)",
            strerror(opal_socket_errno),
            opal_socket_errno));
        return OMPI_ERR_UNREACH;
    }
 
    timeout.tv_sec  = 0;
    timeout.tv_usec = mca_btl_eth_component.eth_echo_timeout;
    if (setsockopt(s, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout)) < 0) {
        BTL_ERROR(("setsockopt(SO_RCVTIMEO): failed: %s (%d)",
            strerror(opal_socket_errno),
            opal_socket_errno));
        return OMPI_ERR_UNREACH;
    }

    /* check each announced address */
    for (i=0; (btl_endpoint->endpoint_addr == NULL) &&
            (i < btl_proc->proc_addr_count); i++) {
    	int try = 0;
        struct mca_btl_eth_addr_t *endpoint_addr = &btl_proc->proc_addrs[i];

        /* check if this remote address is already used by any local BTL module */
        if (endpoint_addr->addr_inuse != 0)
            continue;

#if BTL_ETH_DEBUG_IFLOOKUP
        BTL_OUTPUT(("checking remote proc (%d %d) address %s via if %d", 
            btl_proc->proc_name.jobid,
            btl_proc->proc_name.vpid,
            ether_ntoa(&endpoint_addr->addr_enet),
            btl_eth->eth_ifkindex));
#endif
        while ((btl_endpoint->endpoint_addr == NULL) &&
                (try < mca_btl_eth_component.eth_echo_attempts)) {
            char   buffer[sizeof(struct edp_cmsg_header)];
            struct sockaddr_en sen;
            unsigned int addrlen = sizeof(struct sockaddr_en);
            struct edp_cmsg_header *cm_hdr = (struct edp_cmsg_header *) buffer;

            memset(&sen, 0, sizeof(sen));
            sen.sen_family  = PF_ENET;
            sen.sen_port    = 0;   /* send control message directly to remote kernel */
            sen.sen_ifindex = btl_eth->eth_ifkindex;
            sen.sen_addrlen = ETH_ALEN;
            ETH_ARR2ADDR(sen.sen_addr) = endpoint_addr->addr_enet;
       
            memset(buffer, 0, sizeof(buffer));
            cm_hdr->msg_type = EDP_ECHO_REQUEST;
	
            if (sendto(s, buffer, sizeof(buffer), 0, (struct sockaddr*)&sen,
                    sizeof(sen)) < 0) {
                BTL_ERROR(("sendto(): failed with errno= %s (%d)",
                    strerror(opal_socket_errno),
                    opal_socket_errno));
                return OMPI_ERR_UNREACH;
            }

            if (recvfrom(s, buffer, sizeof(buffer), 0, (struct sockaddr*)&sen,
                    &addrlen) < 0) {
                if (opal_socket_errno == EINTR)
                    continue;
                if (opal_socket_errno != EAGAIN &&
                        opal_socket_errno != EWOULDBLOCK) {
                    BTL_ERROR(("recvfrom(): failed: %s (%d)",
                        strerror(opal_socket_errno),
                        opal_socket_errno));
                    return OMPI_ERR_UNREACH;
                }
                /* the operation timed out - try again */
                ++try;
            } else {
            	/* if we got an echo reply - we're happily done... */
                if (!ETH_ADDR_CMP(sen.sen_addr, ETH_ADDR2ARR(endpoint_addr->addr_enet)) &&
                        ntohs(sen.sen_port) == 0 &&
                        cm_hdr->msg_type == EDP_ECHO_REPLY) {
#if BTL_ETH_DEBUG_IFLOOKUP
                    BTL_OUTPUT(("remote proc (%d %d) address %s IS reachable via if %d", 
                        btl_proc->proc_name.jobid,
                        btl_proc->proc_name.vpid,
                        ether_ntoa(&endpoint_addr->addr_enet),
                        btl_eth->eth_ifkindex));
#endif
                    btl_endpoint->endpoint_addr = endpoint_addr;
                    break;
                }
            }
        }

#if BTL_ETH_DEBUG_IFLOOKUP
        /* debugging output if we didn't receive an answer */
        if (btl_endpoint->endpoint_addr == NULL) {
             BTL_OUTPUT(("remote proc (%d %d) address %s is NOT reachable via if %d", 
                btl_proc->proc_name.jobid,
                btl_proc->proc_name.vpid,
                ether_ntoa(&endpoint_addr->addr_enet),
                btl_eth->eth_ifkindex));
        }
#endif

    }
    
    /* Mark a possibly found endpoint address as 'in use' */
    if( NULL != btl_endpoint->endpoint_addr ) {
        btl_endpoint->endpoint_addr->addr_inuse++;
        return OMPI_SUCCESS;
    }

    return OMPI_ERR_UNREACH;
}

/*
 * Remove an endpoint from the proc array and indicate the address is
 * no longer in use.
 */
                                                                                                                 
int mca_btl_eth_proc_remove(mca_btl_eth_proc_t* btl_proc, mca_btl_base_endpoint_t* btl_endpoint)
{
    size_t i;
    OPAL_THREAD_LOCK(&btl_proc->proc_lock);
    for(i=0; i<btl_proc->proc_endpoint_count; i++) {
        if(btl_proc->proc_endpoints[i] == btl_endpoint) {
            memmove(btl_proc->proc_endpoints+i, btl_proc->proc_endpoints+i+1,
                (btl_proc->proc_endpoint_count-i-1)*sizeof(mca_btl_base_endpoint_t*));
            if(--btl_proc->proc_endpoint_count == 0) {
                OPAL_THREAD_UNLOCK(&btl_proc->proc_lock);
                OBJ_RELEASE(btl_proc);
                return OMPI_SUCCESS;
            }
            /* The endpoint_addr may still be NULL if this enpoint is
               being removed early in the wireup sequence (e.g., if it
               is unreachable by all other procs) */
            if (NULL != btl_endpoint->endpoint_addr) {
                btl_endpoint->endpoint_addr->addr_inuse--;
            }
            break;
        }
    }
    OPAL_THREAD_UNLOCK(&btl_proc->proc_lock);
    return OMPI_SUCCESS;
}

/*
 * Look for an existing ETH process instance based on the globally unique
 * process identifier.
 */
mca_btl_eth_proc_t* mca_btl_eth_proc_lookup(const orte_process_name_t *name)
{
    mca_btl_eth_proc_t* proc = NULL;
    OPAL_THREAD_LOCK(&mca_btl_eth_component.eth_lock);
    opal_hash_table_get_value_uint64(&mca_btl_eth_component.eth_procs, 
                                     orte_util_hash_name(name), (void**)&proc);
    OPAL_THREAD_UNLOCK(&mca_btl_eth_component.eth_lock);
    return proc;
}

/*
 * loop through all available BTLs for one matching the source address
 * of the request.
 */
bool mca_btl_eth_proc_accept(mca_btl_eth_proc_t* btl_proc, struct sockaddr_en* addr, int sd)
{
    size_t i;
    OPAL_THREAD_LOCK(&btl_proc->proc_lock);
    for( i = 0; i < btl_proc->proc_endpoint_count; i++ ) {
        mca_btl_base_endpoint_t* btl_endpoint = btl_proc->proc_endpoints[i];
        /* Check all conditions before going to try to accept the connection. */
        if( memcmp( &btl_endpoint->endpoint_addr->addr_enet,
                    &(((struct sockaddr_en*)addr)->sen_addr),
                    ETH_ALEN ) ) {
            continue;
        }

        if(mca_btl_eth_endpoint_accept(btl_endpoint, addr, sd)) {
            OPAL_THREAD_UNLOCK(&btl_proc->proc_lock);
            return true;
        }
    }
    OPAL_THREAD_UNLOCK(&btl_proc->proc_lock);
    return false;
}

/*
 * convert internal data structure (mca_btl_eth_addr_t) to sockaddr_storage
 *
 */
bool mca_btl_eth_proc_tosocks(mca_btl_eth_addr_t* proc_addr,
                              struct sockaddr_storage* output)
{
    memset(output, 0, sizeof (*output));
    output->ss_family = AF_ENET;
    memcpy(&((struct sockaddr_en*)output)->sen_addr,
           &proc_addr->addr_enet, ETH_ALEN);
    ((struct sockaddr_en*)output)->sen_port = proc_addr->addr_port;
    ((struct sockaddr_en*)output)->sen_addrlen = ETH_ALEN;
    return true;
}

