/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <net/if.h> /* if_nametoindex(...) */
#include <net/ethernet.h>
#include <sys/socket.h>

#include "af_enet.h"

/* these FDs are always open, give them names */
#define FD_STDIN  0
#define FD_STDOUT 1
#define FD_STDERR 2

void print_help(char *exe_name) {
   fprintf(stderr, "usage: %s [remote address] [remote port]\n", exe_name);
   fprintf(stderr, " --buffer-size    -b  Specify size of buffer in bytes.\n");
   fprintf(stderr, " --verbose        -v  Be verbose.\n");
   fprintf(stderr, " --interface      -i  Name of the interface to use.\n");
   fprintf(stderr, " --listen         -l  Set port number to listen on.\n");
   fprintf(stderr, "                      (activates passive mode)\n");
   fprintf(stderr, " --help           -h  Print this message.\n");
}

int main(int argc, char *argv[]) {
   int c;
   int digit_optind = 0;
   int listen_port = 0, sock = 0, if_index = 0;
   struct sockaddr_en address;
   char *if_name = 0;
   int verbose = 0;
   int buffer_size = 4 * 1024;
   void *ibuff = 0, *obuff = 0;
   int ibuff_fill = 0, obuff_fill = 0;
   void *ibuff_pos = 0, *obuff_pos = 0;
   fd_set read_fds, write_fds;
   int res, stdin_open, sock_open;

   /* parse command line options */
   while (1) {
      int this_option_optind = optind ? optind : 1;
      int option_index = 0;
      struct option long_options[] =
         {
            {"listen", 1, 0, 0},
            {"interface", 1, 0, 0},
            {"verbose", 0, 0, 0},
            {"buffer-size", 1, 0, 0},
            {"help", 0, 0, 0},
            {0, 0, 0, 0}
         };
      
      c = getopt_long(argc, argv, "l:i:vb:h", long_options, &option_index);
      if (c == -1)
         break;
      
      switch (c) {

      case 0:
         if (option_index == 0) {
            listen_port = atoi(optarg);
         } else if (option_index == 1) {
            if_name = optarg;
         } else if (option_index == 2) {
            verbose++;
         } else if (option_index == 3) {
            buffer_size = atoi(optarg);
         } else if (option_index == 4) {
            print_help(argv[0]);
            return 0;
         }
         break;
         
      case 'l':
         listen_port = atoi(optarg);
         break;
         
      case 'i':
         if_name = optarg;
         break;
         
      case 'v':
         verbose++;
         break;
         
      case 'b':
         buffer_size = atoi(optarg);
         break;
         
      case 'h':
         print_help(argv[0]);
         return 0;
         
      default:
         printf ("?? internal error: getopt returned character code 0%o ??\n", c);
      }
   }
   
   /* clean address out, is filled in below */
   memset(&address, 0, sizeof(struct sockaddr_en));

   /* read remote address / port */

   if (optind < argc) {
      if (listen_port != 0) {
         fprintf(stderr, "Can't listen AND connect.\n");
         return -1;
      }
      
      if (!ether_aton_r(argv[optind], &address.sen_addr)) {
         fprintf(stderr, "Malformed ethernet address %s\n", argv[optind]);
         return -1;
      } else if (verbose > 0) {
         fprintf(stderr, "Remote ethernet address %s\n", argv[optind]);
      }

      optind++;

      if (optind == argc) {
         fprintf(stderr, "Missing remote port number.\n");
         return -1;
      }

      address.sen_port = htons(atoi(argv[optind]));
      
      if (verbose > 0) {
         fprintf(stderr, "Remote port number %d\n", atoi(argv[optind]));
      }
   } 
   

   /* get interface index */
   if (if_name == 0) {
      /* we need to know where to listen */
      fprintf(stderr, "No interface name given.\n");
      print_help(argv[0]);
      return -1;
   } else {
      if_index = if_nametoindex(if_name);
      
      if (if_index == 0) {
         fprintf(stderr,
                 "Could get index of an interface with name \"%s\"\n",
                 if_name);
         return -1;
      }
   }

   /* set up the connection */
   if (listen_port != 0) {
      /* passive (server) part */

      int result, srv_sock;
      srv_sock = socket(PF_ENET, SOCK_STREAM, EPPROTO_ESP);

      if (srv_sock < 0) {
         perror("socket(...) fehlgeschlagen");
         return -1;
      }

      memset(&address, 0, sizeof(struct sockaddr_en));
      address.sen_family = AF_ENET;
      address.sen_port = htons(listen_port);
      address.sen_ifindex = if_index;
      result = bind(srv_sock,
                    (struct sockaddr*)&address, 
                    sizeof(struct sockaddr_en));
      
      if (result < 0) {
         perror("Could not bind to port");
         return -1;
      }

      if (listen(srv_sock, 1) == -1) {
         perror("listen() failed");
         return -1;
      }

      if (verbose > 0) {
         fprintf(stderr, "Waiting for connection from client...\n");
         fflush(stderr);
      }

      sock = accept(srv_sock, 0, 0);

      if (sock < 0) {
         perror("Could not accept client connection");
         return -1;
      } else if (verbose > 0) {
         fprintf(stderr, "Connected to server.\n");
         fflush(stderr);
      }

      close(srv_sock);

   } else {
      /* active (client) part */
      sock = socket(PF_ENET, SOCK_STREAM, EPPROTO_ESP);

      if (sock < 0) {
         perror("socket(...) fehlgeschlagen");
         return -1;
      }

      address.sen_family = AF_ENET;
      address.sen_ifindex = if_index;
      
      if (verbose > 0) {
         fprintf(stderr, "Trying to connect...\n");
         fflush(stderr);
      }
         
      if (connect(sock, 
                  (struct sockaddr*)(&address),
                  sizeof(struct sockaddr_en)) < 0) {
         perror("Could not connect");
         return -1;
      } else if (verbose > 0) {
         fprintf(stderr, "Connected.\n");
         fflush(stderr);
      }
      
   }

   /* allocate buffer space. ibuff is for stdin, obuff for stdout */
   ibuff = malloc(buffer_size);
   obuff = malloc(buffer_size);
   if ((ibuff == 0) || (obuff == 0)) {
      fprintf(stderr, "Could not allocate buffer space.");
      return -1;
   }

   /* set FDs to non-blocking operation */
   fcntl(FD_STDIN, F_SETFL, O_NONBLOCK);
   fcntl(FD_STDOUT, F_SETFL, O_NONBLOCK);
   fcntl(sock, F_SETFL, O_NONBLOCK);

   /* ok, we start to interfere... 
    * what we do here:
    *
    * (1)
    * - read from stdin to ibuff
    * - if buffer1 is non-empty: write contents to socket
    *
    * (2)
    * - read from socket to obuff
    * - if buffer2 is non-empty: write contents to stdout
    */

   stdin_open = sock_open = 1;
   while (sock_open) {
      /* prepare select() call */
      FD_ZERO(&read_fds);
      if (stdin_open) FD_SET(FD_STDIN, &read_fds);
      if (sock_open) FD_SET(sock, &read_fds);

      FD_ZERO(&write_fds);
      if (obuff_fill > 0) FD_SET(FD_STDOUT, &write_fds);
      if (ibuff_fill > 0) FD_SET(sock, &write_fds);
      
      /* first param?! -> sock will always be the highest FD. good. */
      res = select(sock + 1, &read_fds, &write_fds, 0, 0);

      if (res < 0) {
         perror("Error with call to select(...)");
      } else if (res == 0) {
         /* timeout */
      } else {
         /* we may read or write */
         
         if (FD_ISSET(FD_STDIN, &read_fds) && (ibuff_fill == 0)) {
            /* read to ibuff */
            ibuff_fill = read(FD_STDIN, ibuff, buffer_size);
            if (ibuff_fill <= 0) {
               stdin_open = 0;
               if (verbose > 1) fprintf(stderr, "stdin closed\n");
               fflush(stderr);
            }
            ibuff_pos = ibuff;
         }

         if (FD_ISSET(sock, &write_fds) && (ibuff_fill > 0)) {
            size_t w = write(sock, ibuff_pos, ibuff_fill);
            ibuff_pos += w;
            ibuff_fill -= w;
         }

         if (FD_ISSET(sock, &read_fds) && (obuff_fill == 0)) {
            obuff_fill = read(sock, obuff, buffer_size);
            if (obuff_fill <= 0) {
               sock_open = 0;
               if (verbose > 1) fprintf(stderr, "socket closed\n");
               fflush(stderr);
            }
            obuff_pos = obuff;
         }

         if (FD_ISSET(FD_STDOUT, &write_fds) && (obuff_fill > 0)) {
            size_t w = write(FD_STDOUT, obuff_pos, obuff_fill);
            obuff_pos += w;
            obuff_fill -= w;
         }

      }
   }

   /* flush the buffers */

   while ((ibuff_fill > 0) || (obuff_fill > 0)) {
      FD_ZERO(&write_fds);
      if (obuff_fill > 0) FD_SET(FD_STDOUT, &write_fds);
      if (ibuff_fill > 0) FD_SET(sock, &write_fds);

      if (FD_ISSET(sock, &write_fds) && (ibuff_fill > 0)) {
         size_t w = write(sock, ibuff_pos, ibuff_fill);
         ibuff_pos += w;
         ibuff_fill -= w;
      }
      
      if (FD_ISSET(FD_STDOUT, &write_fds) && (obuff_fill > 0)) {
         size_t w = write(FD_STDOUT, obuff_pos, obuff_fill);
         obuff_pos += w;
         obuff_fill -= w;
      }
   }

   free(ibuff);
   free(obuff);
   close(sock);

   return 0;
}
