/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

/*
 * ENET  An implementation of an Ethernet protocol suite for the LINUX
 *  operating system.  ENET is implemented using the  BSD Socket
 *  interface as the means of communication with the user level.
 *
 *  EDP - ethernet datagram protocol
 *  Implements an unreliable means to send and receive datagrams
 *
 * Version: $Id: enet_dgram.c,v 0.1 2005/xx/xx 10:00:00 rmir Exp $
 *
 * Authors: Mirko Reinhardt, <rmir@hrz.tu-chemnitz.de>
 *
 * Fixes:
 */

#include <linux/version.h> /* kernel version dependency */
#include <linux/net.h>  /* struct socket, struct proto_ops */
#include <net/sock.h>  /* struct sock, basic socket operations */
#include <linux/netdevice.h> /* struct net_device, basic netdevice ops */
#include <linux/skbuff.h> /* socket buffer struct and ops */
#include <asm/ioctls.h>  /* IOCTL constants */
#include <asm/io.h>  /* page_to_phys() and virt_to_phys() */
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "enet_module.h" /* project-global include */
#include "enet_socket.h" /* ENET socket global ops */
#include "enet_dgram.h"

#include "ptable.h"
#include "enet_compat.h"



/*************************************************************************
 * Internal data structures
 ************************************************************************/


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11)


/* edp protocol descriptor */
static struct proto edp_proto = {
	.name     = "EDP",
	.owner    = THIS_MODULE,
	.obj_size = sizeof(struct edp_sock),
};


/* register the edp protocol and create the socket slab cache */
static int edp_prot_init_slab(void)
{
	return proto_register(&edp_proto, 1);
}


/* unregister protocol edp and destroy socket slab cache */
static void edp_prot_destroy_slab(void)
{
	proto_unregister(&edp_proto);
}


/* edp socket allocation function */
static inline struct sock *edp_sk_alloc(NET_PARAM gfp_t priority) {
	return sk_alloc(NET_ARG PF_ENET, priority, &edp_proto);
}


#else /* pre-2.6.10 kernels */


/* edp socket slab-cache */
static kmem_cache_t *edp_sk_cachep = NULL;


/* create the socket slab cache */
static int edp_prot_init_slab(void)
{
	int rc = 0;

	edp_sk_cachep = kmem_cache_create("edp_sock",
	                                  sizeof(struct edp_sock), 0, SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!edp_sk_cachep) {
		PR_ERR("could not create edp socket slab cache");
		rc = -ENOMEM;
	}
	return rc;
}


/* destroy socket slab cache */
static void edp_prot_destroy_slab(void)
{
	if (edp_sk_cachep)
		kmem_cache_destroy(edp_sk_cachep);
}

/* edp socket allocation function */
static inline struct sock *edp_sk_alloc(gfp_t priority) {
	return sk_alloc(PF_ENET, priority, sizeof(struct edp_sock),
	                edp_sk_cachep);
}


#endif


/* counter of EDP sockets alive */
atomic_t edp_skcount;

/* list of all EDP sockets and its lock */
HLIST_HEAD(edp_sklist);
DEFINE_RWLOCK(edp_sklist_lock);


/* port-to-socket hashtable */
struct hlist_head edp_hash[EDP_HTABLE_SIZE];


/* port-to-socket hashtable r/w-lock */
rwlock_t edp_hash_lock = RW_LOCK_UNLOCKED;


/* port range to automatically choose ports from and port rover */
int edp_local_port_range[2] = { 1, 65535 };
int edp_port_rover = 0;

/*
 * Determine whether the given port is free to use
 * or (if rport == 0) the next free port to use
 */
static int edp_get_port(struct sock *sk, int ifindex, unsigned short rport)
{
	struct edp_opt *edp_sk = EDP_SK(sk);
	struct hlist_node *node;
	struct sock *sk2;

	write_lock_bh(&edp_hash_lock);
	if (rport == 0) {
		/* user suppled 0 for port - determine a good free one */
		int best_size_so_far, best, result, i;

		if (edp_port_rover > edp_local_port_range[1] ||
		                edp_port_rover < edp_local_port_range[0])
			edp_port_rover = edp_local_port_range[0];
		best_size_so_far = 32767;
		best = result = edp_port_rover;
		/* look for the next empty or the shortest queue in the hashtable */
		for (i = 0; i < EDP_HTABLE_SIZE; i++, result++) {
			struct hlist_head *list;
			int size;

			list = &edp_hash[result & (EDP_HTABLE_SIZE - 1)];
			if (hlist_empty(list)) {
				if (result > edp_local_port_range[1])
					result = edp_local_port_range[0] +
					         ((result - edp_local_port_range[0]) &
					          (EDP_HTABLE_SIZE - 1));
				goto gotit;
			}
			size = 0;
			sk_for_each_bound(sk2, node, list)
			if (++size >= best_size_so_far)
				goto next;
			best_size_so_far = size;
			best = result;
next:
			;
		}
		/* determine the next free port in the found hashtable queue */
		result = best;
		for (i = 0; i < (1 << 16) / EDP_HTABLE_SIZE; i++, result += EDP_HTABLE_SIZE) {
			if (result > edp_local_port_range[1])
				result = edp_local_port_range[0]
				         + ((result - edp_local_port_range[0]) &
				            (EDP_HTABLE_SIZE - 1));
			if (!edp_port_inuse(result, ifindex))
				break;
		}
		if (i >= (1 << 16) / EDP_HTABLE_SIZE)
			goto fail;
gotit:
		edp_port_rover = rport = result;
	} else {
		/* user supplied a port - check if it can be used */
		sk_for_each_bound(sk2, node, &edp_hash[rport & (EDP_HTABLE_SIZE - 1)]) {
			struct edp_opt *edp_sk2 = EDP_SK(sk2);

			if (edp_sk2->port == rport && sk2 != sk &&
			                (!edp_sk2->ifindex || !ifindex ||
			                 edp_sk2->ifindex == ifindex))
				goto fail;
		}
	}
	/* insert the socket into the hashtable */
	if (sk_unbound(sk)) {
		struct hlist_head *h = &edp_hash[rport & (EDP_HTABLE_SIZE - 1)];

		PR_DEBUG("hashing edp socket %p (port %d)", sk, rport);

		sk_add_bind_node(sk, h);
		edp_sk->port    = rport;
		edp_sk->sport   = htons(rport);
		edp_sk->ifindex = ifindex;
	}
	write_unlock_bh(&edp_hash_lock);
	return 1;

fail:
	write_unlock_bh(&edp_hash_lock);
	return 0;
}


/*
 * Remove the given socket from the port hashtable
 */
static void edp_unhash(struct sock *sk)
{
	struct edp_opt *edp_sk = EDP_SK(sk);

	PR_DEBUG("unhashing edp socket %p", sk);

	write_lock_bh(&edp_hash_lock);
	sk_del_bind_node_init(sk);
	edp_sk->port    = 0;
	edp_sk->sport   = 0;
	edp_sk->ifindex = 0;
	write_unlock_bh(&edp_hash_lock);
}


/*
 * BIND handler
 */
static int sock_edp_bind(struct socket *sock, struct sockaddr *uaddr,
                         int addr_len)
{
	struct sockaddr_en *sen = (struct sockaddr_en*)uaddr;
	struct sock *sk = sock->sk;
	struct net_device *dev = NULL;
	unsigned short port;
	int err = 0;

	// check legality of given address
	if (addr_len < sizeof(*sen))
		return -EINVAL;

	if (sen->sen_family != AF_ENET)
		return -EAFNOSUPPORT;

	/* undo the byte shuffling for ports hashtable */
	port = ntohs(sen->sen_port);

	PR_DEBUG("bind edp socket %p to interface %d port %u",
	         sk, sen->sen_ifindex, port);

	/* bind to specific interface or all interfaces using ifindex == 0 */
	if (sen->sen_ifindex) {
		dev = dev_get_by_index(INIT_NET_ARG sen->sen_ifindex);
		if (!dev) {
			err = -ENODEV;
			goto out;
		}
	}

	/* report an error if the device is down */
	if (dev && !(dev->flags & IFF_UP)) {
		err = -ENETDOWN;
		SK_ERR_REPORT(sk, err);
		goto out_put;
	}

	lock_sock(sk);

	/* report an error if already bound */
	if (sk_bound(sk)) {
		err = -EINVAL;
		goto out_release;
	}

	/* Check the port and bind */
	if (!edp_get_port(sk, sen->sen_ifindex, port)) {
		PR_ERR("failed to assign local port to edp socket %p", sk);
		err = -EADDRINUSE;
		goto out_release;
	}

	PR_DEBUG("bound edp socket %p to interface %s port %u", sk,
	         dev ? dev->name : "[ALL]", EDP_SK(sk)->port);

out_release:
	release_sock(sk);
out_put:
	if (dev)
		dev_put(dev);
out:
	return err;
}


/*
 * CONNECT handler
 */
static int sock_edp_connect(struct socket *sock, struct sockaddr *uaddr,
                            int addr_len, int flags)
{
	struct net_device *dev;
	struct sock *sk = sock->sk;
	struct edp_opt *edp_sk = EDP_SK(sk);
	struct sockaddr_en *sen = (struct sockaddr_en *) uaddr;
	int err = 0;

	PR_DEBUG("connect edp socket %p to host "NETH_FORMAT" port %u via if %d",
	         sk,
	         NETH_HEX(sen->sen_addr),
	         ntohs(sen->sen_port),
	         sen->sen_ifindex);

	if (addr_len < sizeof(*sen))
		return -EINVAL;

	if (sen->sen_family != AF_ENET)
		return -EAFNOSUPPORT;

	dev = dev_get_by_index(INIT_NET_ARG sen->sen_ifindex);
	if (!dev)
		return -ENODEV;

	/* check if device is UP */
	if (!(dev->flags & IFF_UP)) {
		err = -ENETDOWN;
		goto out_put;
	}

	lock_sock(sk);

	if (sk_bound(sk) && edp_sk->ifindex != dev->ifindex) {
		PR_ERR("can't connect edp socket %p via if %d - already bound to if %d",
		       sk, dev->ifindex, edp_sk->ifindex);
		err = -EINVAL;
		goto out_release;
	}

	if (!sk_bound(sk) && !edp_get_port(sk, dev->ifindex, 0)) {
		PR_ERR("failed to assign local port to edp socket %p", sk);
		err = -EAGAIN;
		goto out_release;
	}

	edp_sk->dport = sen->sen_port;
	memcpy(edp_sk->daddr, sen->sen_addr, ETH_ALEN);
	edp_sk->connected = 1;

out_release:
	release_sock(sk);

out_put:
	dev_put(dev);

	return err;
}


/*
 * GETNAME handler - handles getsockname() and getpeername()
 */
static int sock_edp_getname(struct socket *sock, struct sockaddr *uaddr,
                            int *uaddr_len, int peer)
{
	struct net_device *dev;
	struct sock *sk = sock->sk;
	struct edp_opt *edp_sk = EDP_SK(sk);
	struct sockaddr_en *sen = (struct sockaddr_en *) uaddr;
	int err = 0;

	memset(sen, 0, sizeof(*sen));
	sen->sen_family  = AF_ENET;

	lock_sock(sk);

	if (peer) {
		if (!edp_sk->connected) {
			err = -ENOTCONN;
			goto out_release;
		}
		sen->sen_ifindex = edp_sk->ifindex;
		sen->sen_port    = edp_sk->dport;
		sen->sen_addrlen = ETH_ALEN;
		memcpy(sen->sen_addr, edp_sk->daddr, ETH_ALEN);
	} else {
		sen->sen_ifindex = edp_sk->ifindex;
		sen->sen_port    = edp_sk->sport;
		dev = dev_get_by_index(INIT_NET_ARG edp_sk->ifindex);
		if (dev) {
			sen->sen_addrlen = ETH_ALEN;
			memcpy(sen->sen_addr, dev->dev_addr, ETH_ALEN);
			dev_put(dev);
		} else {
			sen->sen_addrlen = 0;
			err = -ENODEV;
		}
	}
	*uaddr_len = sizeof(*sen);

out_release:
	release_sock(sk);

	return err;
}


/*
 * IOCTL handler
 */
static int sock_edp_ioctl(struct socket *sock, unsigned int cmd,
                          unsigned long arg)
{
	struct sock *sk = sock->sk;

	PR_DEBUG("ioctl command %#x  arg %#lx", cmd, arg);

	switch (cmd) {
	case SIOCOUTQ: {
		int amount = atomic_read(&sk->sk_wmem_alloc);
		return put_user(amount, (int __user *)arg);
	}
	case SIOCINQ: {
		struct sk_buff *skb;
		int amount = 0;

		spin_lock_bh(&sk->sk_receive_queue.lock);
		skb = skb_peek(&sk->sk_receive_queue);
		if (skb)
			amount = skb->len;
		spin_unlock_bh(&sk->sk_receive_queue.lock);
		return put_user(amount, (int __user *)arg);
	}
	case SIOCGSTAMP:
		return sock_get_timestamp(sk, (struct timeval __user *)arg);

	default:
		/*
		 * return dev_ioctl(cmd, (void __user *)arg);
		 * dev_ioctl is no longer exported, so:
		 */
		return -ENOIOCTLCMD;

	}
	return 0;
}


/*
 * GETSOCKOPT handler
 */
static int sock_edp_getsockopt(struct socket *sock, int level, int optname,
                               char __user *optval, int __user *optlen)
{
	int len;
	struct sock *sk = sock->sk;
	struct edp_opt *edp_sk = EDP_SK(sk);
	struct net_device *dev;

	PR_DEBUG("getsockopt(%#x)", optname);

	if (level != SOL_EP)
		return -ENOPROTOOPT;

	if (get_user(len, optlen))
		return -EFAULT;

	switch (optname) {
	case EDP_SO_MTU: {
		/* check provided userspace buffer */
		if (len < sizeof(int))
			return -EINVAL;
		len = sizeof(int);
		/* check if the socket is bound */
		if (!sk_bound(sk))
			return -EINVAL;
		/* calculate mtu */
		dev = dev_get_by_index(INIT_NET_ARG edp_sk->ifindex);
		if (dev) {
			unsigned int mtu = dev->mtu;
			mtu -= EDP_HLEN;

			PR_DEBUG("getsockopt determined %s mtu %u, edp mtu %d",
			         dev->name, dev->mtu, mtu);

			dev_put(dev);
			if (put_user(mtu, (int __user *)optval))
				return -EFAULT;
		} else return -ENODEV;
		break;
	}
	case EDP_SO_STATS: {
		struct edp_stats st;
		/* check provided userspace buffer */
		if (len < sizeof(st))
			return -EINVAL;
		len = sizeof(st);
		spin_lock_bh(&sk->sk_receive_queue.lock);
		st = edp_sk->stats;
		memset(&edp_sk->stats, 0, sizeof(st));
		spin_unlock_bh(&sk->sk_receive_queue.lock);
		st.pkt_rx += st.pkt_rx_drop;

		if (copy_to_user(optval, &st, len))
			return -EFAULT;
		break;
	}
	default:
		return -ENOPROTOOPT;
	}

	/* announce size of the returned value to user */
	if (put_user(len, optlen))
		return -EFAULT;
	return 0;
}


/*
 * SENDMSG handler
 */
static int sock_edp_sendmsg(struct kiocb *iocb, struct socket *sock,
                            struct msghdr *msg, size_t len)
{
	struct sock *sk = sock->sk;
	struct edp_opt *edp_sk = EDP_SK(sk);
	struct sockaddr_en *sen = (struct sockaddr_en *)msg->msg_name;
	struct net_device *dev = NULL;
	struct sk_buff *skb;
	struct edp_header *edp_hdr;
	unsigned char *daddr;
	unsigned short dport;
	int ifindex, err = 0;

	/* check whether an address was provided or if the socket is connected */
	if (sen) {
		if (msg->msg_namelen < sizeof(struct sockaddr_en)) {
			err = -EINVAL;
			goto out;
		}
		ifindex = sen->sen_ifindex;
		dport   = sen->sen_port;
		daddr   = sen->sen_addr;
	} else {
		if (!edp_sk->connected) {
			err = -EDESTADDRREQ;
			goto out;
		}
		ifindex = edp_sk->ifindex;
		dport   = edp_sk->dport;
		daddr   = edp_sk->daddr;
	}

	PR_DEBUG("sending edp packet of length %d to "NETH_FORMAT
	         " port %d via if %d",
	         (int)len, NETH_HEX(daddr), ntohs(dport), ifindex);

	// get device for status and message size check
	dev = dev_get_by_index(INIT_NET_ARG ifindex);
	if (!dev) {
		PR_ERR("device %d not found", ifindex);
		err = -ENODEV;
		goto out;
	}

	/* check if device is UP */
	if (!(dev->flags & IFF_UP)) {
		PR_ERR("device %s is DOWN", dev->name);
		err = -ENETDOWN;
		SK_ERR_REPORT(sk, err);
		goto out_put;
	}

	/* autobind socket if not already bound
	 * (first pre-check without the locking overhead) */
	if (!sk_bound(sk)) {
		lock_sock(sk);
		/* check again, after the socket has been locked */
		if (!sk_bound(sk) && !edp_get_port(sk, dev->ifindex, 0)) {
			PR_ERR("failed to assign local port to edp socket %p", sk);
			err = -EAGAIN;
			release_sock(sk);
			goto out_put;
		}
		release_sock(sk);
	}

	/* message size must be less than or equal to device MTU */
	if (len > dev->mtu - EDP_HLEN) {
		PR_ERR("message size %d bigger than protocol mtu %d", (unsigned int)len, dev->mtu);
		err = -EMSGSIZE;
		goto out_put;
	}

	/* Allocate a socket buffer of the needed size from the socket's write
	 * memory buffer - performs sock_hold(sk)
	 */
	skb = sock_alloc_send_skb(sk, LL_RESERVED_SPACE(dev) + EDP_HLEN + len,
	                          msg->msg_flags & MSG_DONTWAIT, &err);
	if (!skb) {
		PR_ERR("failed to allocate socket buffer with error %d", err);
		err = -ENOBUFS;
		goto out_put;
	}

	/* FIXME: Save some space for broken drivers that write a
	 * hard header at transmission time by themselves. PPP is the
	 * notable one here. This should really be fixed at the driver level.
	 */
	skb_reserve(skb, LL_RESERVED_SPACE(dev));
	skb_reset_network_header(skb);

	/* Copy ethernet header into skb (function performs htons(type) ) */
	if (HAS_HARD_HEADER(dev)) {
		if (dev_hard_header(skb, dev, ETH_P_EDP, daddr, NULL, len) < 0) {
			err = -EINVAL;
			goto out_free;
		}
	}

	/* copy EDP header into socket buffer */
	edp_hdr = (struct edp_header*) skb_put(skb, EDP_HLEN);
	edp_hdr->dst_port = dport;
	edp_hdr->src_port = edp_sk->sport;
	edp_hdr->length   = htons(len);

	/* copy the user-supplied data into the socket buffer */
	memcpy_fromiovec(skb_put(skb, len), msg->msg_iov, len);

	/* finish skb */
	skb->dev = dev;
	skb->protocol = __constant_htons(ETH_P_EDP);
	skb->priority = sk->sk_priority;

	/* send skb */
	err = dev_queue_xmit(skb);

	if (err != NET_XMIT_SUCCESS) {
		PR_ERR("failed to enqueue edp socket buffer for sending with errno %d", err);
		if (err > 0)
			err = net_xmit_errno(err);
		if (err != 0) {
			spin_lock(&sk->sk_receive_queue.lock);
			++edp_sk->stats.tx_err;
			spin_unlock(&sk->sk_receive_queue.lock);
			goto out_put;
		}
	}

	/* count stats */
	spin_lock(&sk->sk_receive_queue.lock);
	++edp_sk->stats.pkt_tx;
	spin_unlock(&sk->sk_receive_queue.lock);

	PR_DEBUG("sent edp frame of length %d type %#x (sk refcnt %d)",
	         skb->len, ntohs(skb->protocol), atomic_read(&sk->sk_refcnt));

	dev_put(dev);

	/* nearly all send*() implementations return the bytes just sent */
	return len;


out_free:
	kfree_skb(skb);
out_put:
	dev_put(dev);
out:
	return err;
}


/*
 * SENDMSG handler (zero-copy)
 */
static int sock_edp_sendmsg_zcpy(struct kiocb *iocb, struct socket *sock,
                                 struct msghdr *msg, size_t len)
{
	struct sock *sk = sock->sk;
	struct edp_opt *edp_sk = EDP_SK(sk);
	struct sockaddr_en *sen = (struct sockaddr_en *)msg->msg_name;
	struct net_device *dev = NULL;
	struct sk_buff *skb;
	struct edp_header *edp_hdr;
	unsigned char *daddr;
	unsigned short dport;
	int ifindex, err = 0;
	int data_len;
	struct iovec *iov;

	/* check whether an address was provided or if the socket is connected */
	if (sen) {
		if (msg->msg_namelen < sizeof(struct sockaddr_en)) {
			err = -EINVAL;
			goto out;
		}
		ifindex = sen->sen_ifindex;
		dport   = sen->sen_port;
		daddr   = sen->sen_addr;
	} else {
		if (!edp_sk->connected) {
			err = -EDESTADDRREQ;
			goto out;
		}
		ifindex = edp_sk->ifindex;
		dport   = edp_sk->dport;
		daddr   = edp_sk->daddr;
	}

	PR_DEBUG("sending edp packet of length %d to "NETH_FORMAT"-%d via if %d",
	         (int)len, NETH_HEX(daddr), ntohs(dport), ifindex);

	// get device for status and message size check
	dev = dev_get_by_index(INIT_NET_ARG ifindex);
	if (!dev) {
		PR_ERR("device %d not found", ifindex);
		err = -ENODEV;
		goto out;
	}

	/* autobind socket if not already bound
	 * (first pre-check without the locking overhead) */
	if (!sk_bound(sk)) {
		lock_sock(sk);
		/* check again, after the socket has been locked */
		if (!sk_bound(sk) && !edp_get_port(sk, dev->ifindex, 0)) {
			PR_ERR("failed to assign local port to edp socket %p", sk);
			err = -EAGAIN;
			release_sock(sk);
			goto out_put;
		}
		release_sock(sk);
	}

	/* check if device is UP */
	if (!(dev->flags & IFF_UP)) {
		PR_ERR("device %s is DOWN", dev->name);
		err = -ENETDOWN;
		SK_ERR_REPORT(sk, err);
		goto out_put;
	}

	/* message size must be less than or equal to device MTU */
	if (len > dev->mtu - EDP_HLEN) {
		PR_ERR("message size %d bigger than protocol mtu %d",
		       (unsigned int)len, dev->mtu);
		err = -EMSGSIZE;
		goto out_put;
	}

	/* we can only allow as much iovecs as fragments to an skb */
	if (msg->msg_iovlen > MAX_SKB_FRAGS) {
		PR_ERR("io vector count %d bigger than maximum fragment count %ld",
		       (int)msg->msg_iovlen, MAX_SKB_FRAGS);
		err = -EMSGSIZE;
		goto out_put;
	}

	/* Allocate a socket buffer of the needed size from the socket's write
	 * memory buffer - performs sock_hold(sk)
	 */
	skb = sock_alloc_send_skb(sk, LL_RESERVED_SPACE(dev) + EDP_HLEN,
	                          msg->msg_flags & MSG_DONTWAIT, &err);
	if (!skb) {
		PR_ERR("failed to allocate socket buffer with error %d", err);
		err = -ENOBUFS;
		goto out_put;
	}

	/* FIXME: Save some space for broken drivers that write a
	 * hard header at transmission time by themselves. PPP is the
	 * notable one here. This should really be fixed at the driver level.
	 */
	skb_reserve(skb, LL_RESERVED_SPACE(dev));
	skb_reset_network_header(skb);

	/* Copy ethernet header into skb (function performs htons(type) ) */
	if (HAS_HARD_HEADER(dev)) {
		if (dev_hard_header(skb, dev, ETH_P_EDP, daddr, NULL, len) < 0) {
			err = -EINVAL;
			goto out_free;
		}
	}

	/* copy EDP header into socket buffer */
	edp_hdr = (struct edp_header*) skb_put(skb, EDP_HLEN);
	edp_hdr->dst_port = dport;
	edp_hdr->src_port = edp_sk->sport;
	edp_hdr->length   = htons(len);

	/*
	 * ZERO-COPY approach for data sending.
	 *
	 * The address of the user data buffer is written to the socket buffer as fragment,
	 * so the NIC can get the data via DMA directly from that memory region.
	 *
	 * The kernel will automatically create a linear copy of the skb (inside function
	 * dev_queue_xmit()) if the NIC is not capable of scatter-gather I/O or any of the
	 * fragments is located in high memory unaccessible to the NIC's DMA controller.
	 */

	data_len = len;
	iov      = msg->msg_iov;

	while (data_len > 0) {
		struct page *frag_page = NULL;
		int frag_off, frag_size;

		/* check for maximum fragment count */
		if (skb_shinfo(skb)->nr_frags >= MAX_SKB_FRAGS) {
			PR_ERR("too many fragments (%d/%ld) to a single skb",
			       skb_shinfo(skb)->nr_frags + 1, MAX_SKB_FRAGS);
			err = -EMSGSIZE;
			goto out_free;
		}

		/* don't waste fragments for empty io vectors */
		if (!iov->iov_len) {
			++iov;
			continue;
		}

		/* Determine the struct page for the given userspace virtual address */
		frag_page = lookup_page(current->mm, (unsigned long)iov->iov_base);

		if (!frag_page) {
			PR_ERR("no page for virtual address %p found", iov->iov_base);
			err = -EFAULT;
			goto out_free;
		}

		/* Mark page frame as "in use" */
		get_page(frag_page);

		/* Determine the fragment values and add the page to the socket buffer as fragment */
		frag_off  = offset_in_page((unsigned long)iov->iov_base);
		frag_size = min_t(unsigned int, iov->iov_len, PAGE_SIZE - frag_off);

		/* Add the page to the socket buffer as fragment */
		skb_fill_page_desc(skb, skb_shinfo(skb)->nr_frags, frag_page, frag_off, frag_size);
		skb->len += frag_size;
		skb->data_len += frag_size;

#if ENET_DEBUG
		PR_DEBUG("iov %p-%d -> page %p (phys %#x off %d len %d)",
		         iov->iov_base, (unsigned int)iov->iov_len,
		         frag_page, (unsigned int)page_to_phys(frag_page),
		         frag_off, frag_size);

		if (!(dev->features & NETIF_F_SG) ||
		                (!(dev->features & NETIF_F_HIGHDMA) && PageHighMem(frag_page)))
			PR_DEBUG("fragment will be copied [ %s%s]", !(dev->features & NETIF_F_SG) ? "NO SG-DMA " : "",
			         (!(dev->features & NETIF_F_HIGHDMA) && PageHighMem(frag_page)) ? "ILLEGAL HIGHDMA " : "");

		/* Imitation of __skb_linearize(), skb_copy_bits() */
		{
			skb_frag_t *frag  = &skb_shinfo(skb)->frags[skb_shinfo(skb)->nr_frags-1];
			u8         *vaddr = kmap_skb_frag(frag);
			unsigned long phys_addr = virt_to_phys(vaddr);
			unsigned long page_addr = page_to_phys(frag->page);
			if (phys_addr != page_addr) {
				PR_DEBUG("wrong struct page in skb: page_to_phys[%#lx]  paddr[%#lx]",
				         page_addr, phys_addr);
			}
			kunmap_skb_frag(vaddr);
		}
#endif

		/* reduce the amount of data to be processed */
		data_len -= frag_size;

		/* if the data length of the current io vector exceeds a page,
		 * another fragment has to be appended... */
		iov->iov_base += frag_size;
		iov->iov_len  -= frag_size;

		/* if the current io vector is covered, proceed to the next one */
		if (iov->iov_len == 0)
			++iov;
	}

	/* finish skb */
	skb->dev = dev;
	skb->protocol = __constant_htons(ETH_P_EDP);
	skb->priority = sk->sk_priority;

	/* send skb */
	err = dev_queue_xmit(skb);

	if (err != NET_XMIT_SUCCESS) {
		PR_ERR("failed to enqueue edp socket buffer for sending with errno %d", err);
		if (err > 0)
			err = net_xmit_errno(err);
		if (err != 0) {
			spin_lock(&sk->sk_receive_queue.lock);
			++edp_sk->stats.tx_err;
			spin_unlock(&sk->sk_receive_queue.lock);
			goto out_put;
		}
	}

	/* count stats */
	spin_lock(&sk->sk_receive_queue.lock);
	++edp_sk->stats.pkt_tx;
	spin_unlock(&sk->sk_receive_queue.lock);

	PR_DEBUG("sent edp frame of length %d type %#x (sk refcnt %d)", skb->len, ntohs(skb->protocol), atomic_read(&sk->sk_refcnt));

	dev_put(dev);

	/* nearly all send*() implementations return the bytes just sent */
	return len;


out_free:
	kfree_skb(skb);
out_put:
	dev_put(dev);
out:
	return err;
}


/*
 * RECVMSG handler
 * Pulls a packet from the receive queue and hands it to the user.
 * Blocks if necessary.
 */
static int sock_edp_recvmsg(struct kiocb *iocb, struct socket *sock,
                            struct msghdr *msg, size_t len, int flags)
{
	struct sock *sk = sock->sk;
	struct sk_buff *skb;
	int copied, err;

	printk("htor: sock_edp_recvmsg\n");
	err = -EINVAL;
	if (flags & ~(MSG_PEEK | MSG_DONTWAIT | MSG_TRUNC | MSG_CMSG_COMPAT))
		goto out;

	/*
	 * Call the generic datagram receiver. This handles all sorts
	 * of horrible races and re-entrancy so we can forget about it
	 * in the protocol layers.
	 *
	 * Now it will return ENETDOWN, if device has just gone down,
	 * but then it will block.
	 */
	skb = skb_recv_datagram(sk, flags, flags & MSG_DONTWAIT, &err);

	if (skb == NULL)
		goto out;

	printk("after skb_recv_datagram()\n");

	/*
	 * You lose any data beyond the buffer you gave. If it worries a
	 * user program they can ask the device for its MTU anyway.
	 */
	copied = skb->len;
	if (copied > len) {
		copied = len;
		msg->msg_flags |= MSG_TRUNC;
	}

	/* copy data to the user space buffer the io vector points to */
	err = skb_copy_datagram_iovec(skb, 0, msg->msg_iov, copied);
	if (err)
		goto out_free;
	printk("after skb_copy_datagram_iovec()\n");

	sock_recv_timestamp(msg, sk, skb);

	/* report the sender's address to the user */
	if (msg->msg_name) {
		struct edp_header *edp_hdr = EDP_HDR(skb);
		struct sockaddr_en *sen = (struct sockaddr_en *) msg->msg_name;

		if (msg->msg_namelen >= sizeof(struct sockaddr_en)) {
			sen->sen_family   = PF_ENET;
			sen->sen_port     = edp_hdr->src_port;
			sen->sen_ifindex  = EDP_SKB_CB(skb)->ifindex;
			sen->sen_addrlen  = EDP_SKB_CB(skb)->addrlen;
			memcpy(sen->sen_addr, EDP_SKB_CB(skb)->addr,
			       EDP_SKB_CB(skb)->addrlen);

			msg->msg_namelen = sizeof(struct sockaddr_en);
		} else msg->msg_namelen = 0;
	}

	printk("after all %i :)\n", err);

	/*
	 * Free or return the buffer as appropriate. Again this
	 * hides all the races and re-entrancy issues from us.
	 */
	err = copied;
	printk("copied %i :)\n", err);

out_free:
	skb_free_datagram(sk, skb);
out:
	return err;
}


/*
 * Release an EDP socket
 */
static int sock_edp_release(struct socket *sock)
{
	struct sock *sk = sock->sk;
	struct edp_opt *edp_sk;

	if (!sk)
		return 0;

	edp_sk = EDP_SK(sk);

	// remove socket from ports hashtable
	if (sk_bound(sk)) {
		edp_unhash(sk);
	}

	/* remove the socket from the sockets list */
	write_lock_bh(&edp_sklist_lock);
	sk_del_node_init(sk);  /* performs __sock_put(sk) */
	write_unlock_bh(&edp_sklist_lock);


	/* now the socket is dead. No more input will appear */
	sock_orphan(sk);
	sock->sk = NULL;

	// purge queues
	skb_queue_purge(&sk->sk_receive_queue);

	PR_DEBUG("edp socket %p released, refcount %d", sk, atomic_read(&sk->sk_refcnt));

	/* release the socket */
	sock_put(sk);
	return 0;
}


/*
 * Destroy a sock structure associated with an EDP socket
 */
static void sock_edp_destruct(struct sock *sk)
{
	WARN_ON(atomic_read(&sk->sk_rmem_alloc));
	WARN_ON(atomic_read(&sk->sk_wmem_alloc));

	if (!sock_flag(sk, SOCK_DEAD)) {
		PR_ERR("attempt to release alive edp socket: %p", sk);
		return;
	}

	atomic_dec(&edp_skcount);

	PR_DEBUG("edp socket %p is free, %d are alive", sk, atomic_read(&edp_skcount));
}



/* EDP socket operations */
static struct proto_ops edp_ops[] = {
	{ /* standard EDP protocol operations */
		.family     = PF_ENET,
		.owner      = THIS_MODULE,
		.release    = sock_edp_release,
		.bind       = sock_edp_bind,
		.connect    = sock_edp_connect,
		.socketpair = sock_no_socketpair,
		.accept     = sock_no_accept,
		.getname    = sock_edp_getname,
		.poll       = sock_no_poll,
		.ioctl      = sock_edp_ioctl,
		.listen     = sock_no_listen,
		.shutdown   = sock_no_shutdown,
		.setsockopt = sock_no_setsockopt,
		.getsockopt = sock_edp_getsockopt,
		.sendmsg    = sock_edp_sendmsg,
		.recvmsg    = sock_edp_recvmsg,
		.mmap       = sock_no_mmap,
		.sendpage   = sock_no_sendpage,
	},
	{ /* EDP zero-copy protocol operations */
		.family     = PF_ENET,
		.owner      = THIS_MODULE,
		.release    = sock_edp_release,
		.bind       = sock_edp_bind,
		.connect    = sock_edp_connect,
		.socketpair = sock_no_socketpair,
		.accept     = sock_no_accept,
		.getname    = sock_edp_getname,
		.poll       = sock_no_poll,
		.ioctl      = sock_edp_ioctl,
		.listen     = sock_no_listen,
		.shutdown   = sock_no_shutdown,
		.setsockopt = sock_no_setsockopt,
		.getsockopt = sock_edp_getsockopt,
		.sendmsg    = sock_edp_sendmsg_zcpy,
		.recvmsg    = sock_edp_recvmsg,
		.mmap       = sock_no_mmap,
		.sendpage   = sock_no_sendpage,
	},
};

#define EDP_PROTOCOL_COUNT (sizeof(edp_ops) / sizeof(struct proto_ops))


/*
 * Create an EDP socket
 */
struct sock *sock_edp_create(NET_PARAM struct socket *sock, int protocol, int *err)
{
	struct sock *sk = NULL;

	if (protocol < 0 || protocol >= EDP_PROTOCOL_COUNT) {
		*err = -EPROTONOSUPPORT;
		goto out;
	}

	/* allocate edp socket structure */
	sk = edp_sk_alloc(NET_ARG GFP_KERNEL);
	if (!sk) {
		PR_ERR("Could not allocate edp kernel socket structure");
		*err = -ENOMEM;
		goto out;
	}

	/* init kernel sock structure */
	sock_init_data(sock, sk);
	/* the  sock->sk_owner  is never used and vanished in 2.6.13 */
	/* sk_set_owner(sk, THIS_MODULE); */
	sk->sk_family   = PF_ENET;
	sk->sk_protocol = protocol;

	// set operations to use with this socket (for bind, send, recv, ...)
	sock->ops = &edp_ops[protocol];

	// set kernel sock destructor
	sk->sk_destruct = sock_edp_destruct;

	atomic_inc(&edp_skcount);

	/* insert the socket into the sockets list */
	write_lock_bh(&edp_sklist_lock);
	sk_add_node(sk, &edp_sklist);  /* performs sock_hold(sk) */
	write_unlock_bh(&edp_sklist_lock);

	PR_DEBUG("edp socket %p created, %d are alive", sk, atomic_read(&edp_skcount));
out:
	return sk;
}



/*
 * EDP control message processing function
 */
static void edp_rcv_cmsg(struct sk_buff *src_skb, struct net_device *src_dev)
{
	struct sk_buff         *dst_skb;
	struct ethhdr          *src_eth_hdr,  *dst_eth_hdr;
	struct edp_header      *src_edp_hdr,  *dst_edp_hdr;
	struct edp_cmsg_header *src_cmsg_hdr, *dst_cmsg_hdr;
	int buf_size, rc;

	/* prepare received control message for processing */
	if (!pskb_may_pull(src_skb, EDP_CMSG_HLEN)) {
		PR_ERR("dropped too short edp control message frame");
		return;
	}

	/* set header pointers */
	src_eth_hdr    = (struct ethhdr *) skb_mac_header(src_skb);
	src_edp_hdr    = (struct edp_header *) skb_network_header(src_skb);
	src_cmsg_hdr   = (struct edp_cmsg_header *) src_skb->data;

	/* pull control message header */
	__skb_pull(src_skb, sizeof(struct edp_cmsg_header));

	/* process control message */
	switch (src_cmsg_hdr->msg_type) {
	case EDP_ECHO_REQUEST:
		PR_DEBUG("received edp echo request with %d data bytes", src_skb->len);
		/* prepare reply socket buffer */
		buf_size = LL_RESERVED_SPACE(src_dev) + EDP_HLEN + EDP_CMSG_HLEN + src_skb->len;
		dst_skb = alloc_skb(buf_size, GFP_ATOMIC);
		skb_reserve(dst_skb, LL_RESERVED_SPACE(src_dev));
		skb_reset_network_header(dst_skb);
		/* fill reply ethernet header */
		EDP_MAC_HDR(dst_skb) = skb_push(dst_skb, ETH_HLEN);
		dst_eth_hdr      = (struct ethhdr *) skb_mac_header(dst_skb);
		memcpy(dst_eth_hdr->h_dest, src_eth_hdr->h_source, ETH_HLEN);
		memcpy(dst_eth_hdr->h_source, src_eth_hdr->h_dest, ETH_HLEN);
		dst_eth_hdr->h_proto = src_eth_hdr->h_proto;
		/* fill reply EDP header */
		dst_edp_hdr = (struct edp_header *) skb_put(dst_skb, EDP_HLEN);
		dst_edp_hdr->dst_port = src_edp_hdr->src_port;
		dst_edp_hdr->src_port = src_edp_hdr->dst_port;
		dst_edp_hdr->length   = src_edp_hdr->length;
		/* fill reply EDP control message header */
		dst_cmsg_hdr = (struct edp_cmsg_header *) skb_put(dst_skb, EDP_CMSG_HLEN);
		dst_cmsg_hdr->msg_type = EDP_ECHO_REPLY;
		/* copy EDP control message data */
		memcpy(skb_put(dst_skb, src_skb->len), src_skb->data, src_skb->len);

		/* finish reply skb */
		dst_skb->dev = src_dev;
		dst_skb->protocol = __constant_htons(ETH_P_EDP);

		/* send reply skb */
		rc = dev_queue_xmit(dst_skb);
		if (rc != NET_XMIT_SUCCESS) {
			PR_ERR("failed to enqueue edp echo reply for sending with errno %d", rc);
			break;
		}
		PR_DEBUG("sent edp echo reply len %d (%d) to "NETH_FORMAT" port %d",
		         dst_skb->len,
		         ntohs(dst_edp_hdr->length),
		         NETH_HEX(dst_eth_hdr->h_dest),
		         ntohs(dst_edp_hdr->dst_port));
		break;

	default:
		PR_ERR("dropped illegal edp control message of type %d", src_cmsg_hdr->msg_type);
	}
}


/*
 * Data receiving kernel hook function
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,13)
static int edp_rcv_hook(struct sk_buff *skb, struct net_device *dev,
                        struct packet_type *pt, struct net_device *orig_dev)
#else
static int edp_rcv_hook(struct sk_buff *skb, struct net_device *dev,
                        struct packet_type *pt)
#endif
{
	struct sock *sk = NULL;
	struct edp_opt *edp_sk = NULL;
	struct edp_header *edp_hdr;
	unsigned short edp_len;
	unsigned short edp_port;
	int skb_len;
	struct hlist_node *node;
#ifdef DEBUG
	struct timeval stamp;
	skb_get_timestamp(skb, &stamp);
	PR_DEBUG("received edp frame on %s of length %d type %#x (ts %ld.%ld sh %d cl %d)",
	         dev->name, skb->len, ntohs(skb->protocol),
	         (long)stamp.tv_sec, (long)stamp.tv_usec,
	         skb_shared(skb), skb_cloned(skb));
#endif
	/* we dont want any packets of other protocols or for other hosts */
	if (skb->protocol != __constant_htons(ETH_P_EDP) ||
	                (skb->pkt_type != PACKET_HOST &&
	                 skb->pkt_type != PACKET_BROADCAST &&
	                 skb->pkt_type != PACKET_MULTICAST)) {
		PR_DEBUG("dropped edp frame for foreign protocol or destination host");
		goto out_free;
	}

	/* check if skb is shared and get a private header copy,
	 * because we will change and enqueue it */
	if ((skb = skb_share_check(skb, GFP_ATOMIC)) == NULL) {
		PR_DEBUG("could not allocate memory to clone a shared edp socket buffer");
		goto out_nomem;
	}

	/* check if there is enough data for at least the EDP header */
	if (!pskb_may_pull(skb, EDP_HLEN)) {
		PR_DEBUG("dropped too short edp frame");
		goto out_free;
	}

	/* pull the EDP header */
	__skb_pull(skb, EDP_HLEN);

	/* inspect EDP header */
	edp_hdr  = EDP_HDR(skb);
	edp_len  = ntohs(edp_hdr->length);
	edp_port = ntohs(edp_hdr->dst_port);

	PR_DEBUG("edp packet header dst %d src %d len %d",
	         edp_port, ntohs(edp_hdr->src_port), edp_len);

	/* trim the packet to the original length if the underlying medium
	 * padded it out */
	if (skb->len > edp_len) {
		PR_DEBUG("trimming received edp packet to %u bytes", edp_len);
		__pskb_trim(skb, edp_len);
	}

	/* if destination port == 0, a control message has been received */
	if (!edp_port) {
		edp_rcv_cmsg(skb, dev);
		goto out_free;
	}

	/* determine the socket from interface and port of received skb -
	 * performs sock_hold(sk) */
	read_lock(&edp_hash_lock);
	EDP_HTABLE_SK_GET(sk, node, edp_hash, EDP_HTABLE_SIZE, dev->ifindex, edp_port);
	read_unlock(&edp_hash_lock);
	if (!sk) {
		PR_DEBUG("no edp socket bound to interface %s port %d",
		         dev->name, edp_port);
		goto out_free;
	}

	edp_sk = EDP_SK(sk);

	/* manually queue received skb to the socket's receive queue to avoid any overhead */
	if (atomic_read(&sk->sk_rmem_alloc) + skb->truesize >= (unsigned)sk->sk_rcvbuf) {
		PR_DEBUG("edp socket %p out of receive memory - dropping packet", sk);
		goto out_dropped;
	}

	/* store some information about the skb in its control data area
	 * because the device that is needed to retrieve the information
	 * might be gone later */
	EDP_SKB_CB(skb)->ifindex = dev->ifindex;
	EDP_SKB_CB(skb)->addrlen = 0;

	if (HAS_HARD_HEADER_PARSE(dev))
		EDP_SKB_CB(skb)->addrlen = dev_parse_header(skb, EDP_SKB_CB(skb)->addr);

	/* hand the socket buffer from the device to the socket */
	skb->dev = NULL;
	skb_set_owner_r(skb, sk);

	/* cache skb length, because once the skb has been added to the queue
	 * it could be pulled away from under our fingers any time by other
	 * threads
	 */
	skb_len = skb->len;
	spin_lock(&sk->sk_receive_queue.lock);
	++edp_sk->stats.pkt_rx;
	__skb_queue_tail(&sk->sk_receive_queue, skb);
	spin_unlock(&sk->sk_receive_queue.lock);
	sk->sk_data_ready(sk, skb_len);

	sock_put(sk);

	PR_DEBUG("enqueued received edp frame to edp socket %p", sk);

	return 0;

out_dropped:
	sock_put(sk);

	spin_lock(&sk->sk_receive_queue.lock);
	++edp_sk->stats.pkt_rx_drop;
	spin_unlock(&sk->sk_receive_queue.lock);
out_free:
	kfree_skb(skb);
out_nomem:
	return 0;
}


/*
 * Netdevice event notifier
 */
static int edp_dev_notifier(struct notifier_block *this, unsigned long msg,
                            void *data)
{
	struct sock *sk;
	struct net_device *dev = (struct net_device*)data;
	struct hlist_node *node;
	int bucket;

	read_lock(&edp_hash_lock);
	HTABLE_SK_FOR_EACH_BOUND(sk, node, bucket, edp_hash, EDP_HTABLE_SIZE) {
		struct edp_opt *edp_sk = EDP_SK(sk);

		switch (msg) {
			// notify sockets of dead devices
		case NETDEV_UNREGISTER:
		case NETDEV_DOWN:
			if (dev->ifindex == edp_sk->ifindex) {
				if (sk_bound(sk)) {
					// remove socket from hashtable
					edp_unhash(sk);
					SK_ERR_REPORT(sk, -ENETDOWN);
				}
			}
			break;
		}
	}

	read_unlock(&edp_hash_lock);
	return NOTIFY_DONE;
}



/*
 * Kernel hook data structures
 */
struct packet_type edp_receive_hook = {
	.type  = __constant_htons(ETH_P_EDP), /**< enter the packet type hashtable */
	.dev  = NULL,    /**< get packets from all devices */
	.func  = edp_rcv_hook,   /**< packet receiving handler function */
	.af_packet_priv = NULL,    /**< private data */
};

struct notifier_block edp_netdev_notifier = {
	.notifier_call = edp_dev_notifier,  /**< netdevice event notifier */
};



/* ProcFS operations */
#ifdef CONFIG_PROC_FS

static inline struct sock *edp_seq_idx(loff_t off) {
	struct sock *sk;
	struct hlist_node *node;

	sk_for_each(sk, node, &edp_sklist) {
		if (!off--)
			return sk;
	}
	return NULL;
}


static void *edp_seq_start(struct seq_file *seq, loff_t *pos)
__acquires(edp_sklist_lock)
{
	read_lock(&edp_sklist_lock);
	return *pos ? edp_seq_idx(*pos - 1) : SEQ_START_TOKEN;
}


static void *edp_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	++*pos;
	return (v == SEQ_START_TOKEN)
	       ? sk_head(&edp_sklist)
	       : sk_next((struct sock*)v) ;
}


static void edp_seq_stop(struct seq_file *seq, void *v)
__releases(edp_sklist_lock)
{
	read_unlock(&edp_sklist_lock);
}


static int edp_seq_show(struct seq_file *seq, void *v)
{
	if (v == SEQ_START_TOKEN)
		seq_puts(seq, "sk       RefCnt Port   Iface Rmem   User   Inode"
		         "  TX     TX ERR RX     Dropped\n");
	else {
		struct sock *sk = v;
		const struct edp_opt *edp_sk = EDP_SK(sk);

		seq_printf(seq,
		           "%p %-6d %-6d %-5d %-6u %-6u %-6lu %-6u %-6u %-6u %-6u\n",
		           sk,
		           atomic_read(&sk->sk_refcnt),
		           edp_sk->port,
		           edp_sk->ifindex,
		           atomic_read(&sk->sk_rmem_alloc),
		           sock_i_uid(sk),
		           sock_i_ino(sk),
		           edp_sk->stats.pkt_tx,
		           edp_sk->stats.tx_err,
		           edp_sk->stats.pkt_rx,
		           edp_sk->stats.pkt_rx_drop
		          );
	}

	return 0;
}


static struct seq_operations edp_seq_ops = {
	.start = edp_seq_start,
	.next = edp_seq_next,
	.stop = edp_seq_stop,
	.show = edp_seq_show,
};


static int edp_seq_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &edp_seq_ops);
}


struct file_operations edp_proc_fops = {
	.owner  = THIS_MODULE,
	.open  = edp_seq_open,
	.read  = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release,
};

#endif


/* init the edp protocol */
int edp_prot_init(void)
{
	int rc = 0;
	struct sk_buff *dummy_skb;

	/* edp skb control data size check */
	if (sizeof(struct edp_skb_cb) > sizeof(dummy_skb->cb)) {
		PR_ERR("edp skb control data structure too big (%d/%d bytes)",
		       (int)sizeof(struct edp_skb_cb),
		       (int)sizeof(dummy_skb->cb));
		rc = -EFAULT;
		goto out;
	}

	rc = edp_prot_init_slab();

out:
	return rc;
}


/* destroy the edp protocol */
void edp_prot_destroy(void)
{
	edp_prot_destroy_slab();
}
