/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

/*
 * ENET  An implementation of an Ethernet protocol suite for the LINUX
 *  operating system.  ENET is implemented using the  BSD Socket
 *  interface as the means of communication with the user level.
 *
 *  Module management functions and kernel hook datastructures
 *
 * Version: $Id: enet_module.c,v 0.1 2005/xx/xx 10:00:00 rmir Exp $
 *
 * Authors: Mirko Reinhardt, <rmir@hrz.tu-chemnitz.de>
 *
 * Fixes:
 */

#include <linux/version.h> /* kernel version dependent compiling */
#include <linux/types.h>
#include <linux/socket.h> /* struct socket */
#include <linux/netdevice.h> /* struct netdevice */
// #include <linux/kmod.h>
#include <linux/skbuff.h> /* basic socket buffer operations */
#include <net/sock.h>  /* struct sock */
#include <linux/errno.h>
#include <asm/system.h>
#include <linux/proc_fs.h>
#include <linux/module.h>
#include <linux/init.h>

#include "enet_module.h"
#include "enet_socket.h"
#include "enet_dgram.h"
#include "enet_stream.h"
#include "enet_stream_cong.h"
#include "enet_compat.h"

/* *****************************************************************************
 *
 * Kernel hook data structures
 *
 */

/* Protocol family operations for the new PF_ENET */
static struct net_proto_family enet_family_ops = {
	.family = PF_ENET,
	.create = sock_enet_create,
	.owner = THIS_MODULE,
};

static unsigned char pf_installed = 0;


enum __prot_states {
	PROT_DOWN               = 0,
	PROT_INITIALIZED        = 1,
	PROT_RECVHOOK_INSTALLED = 2,
	PROT_NOTIFIER_INSTALLED = 3,
	PROT_PROCFS_INSTALLED   = 4,
	PROT_SYSCTL_INSTALLED   = 5
};

#define PROTOCOL_STATE(state, val) ((state) >= (val))

/* implemented ENET protocols */
struct enet_protocol {
	char                   *name;
	int (*init)(void);
	void (*destroy)(void);
	struct packet_type     *receive_hook;
	struct notifier_block  *netdev_notifier;
	struct ctl_table       *sysctl_table;
	struct ctl_table_header *sysctl_table_header;
#ifdef CONFIG_PROC_FS
	struct file_operations *proc_fops;
#endif
	unsigned char          state;
};

static struct enet_protocol enet_protocols[] = {
	{ /* EDP - Ethernet Datagram Protocol */
		.name            = "edp",
		.init            = edp_prot_init,
		.destroy         = edp_prot_destroy,
		.receive_hook    = &edp_receive_hook,
		.netdev_notifier = &edp_netdev_notifier,
		.sysctl_table    = NULL,
		.sysctl_table_header = NULL,
#ifdef CONFIG_PROC_FS
		.proc_fops       = &edp_proc_fops,
#endif
		.state           = PROT_DOWN,
	},
	{ /* ESP - Ethernet Streaming Protocol */
		.name            = "esp",
		.init            = esp_prot_init,
		.destroy         = esp_prot_destroy,
		.receive_hook    = &esp_receive_hook,
		.netdev_notifier = &esp_netdev_notifier,
		.sysctl_table    = esp_ctl_table,
		.sysctl_table_header = NULL,
#ifdef CONFIG_PROC_FS
		.proc_fops       = &esp_proc_fops,
#endif
		.state           = PROT_DOWN,
	},
};

#define ENET_PROTOCOL_COUNT (sizeof(enet_protocols) / sizeof(struct enet_protocol))

/*
 * Kernel Module management functions
 */

/*
 * Cleanup installed ressources and hooks
 */
static void module_cleanup(void)
{
	int i;
	int err;

	/* unregister all installed protocols */
	for (i = 0; i < ENET_PROTOCOL_COUNT; i++) {
		struct enet_protocol *prot = &enet_protocols[i];

		if (PROTOCOL_STATE(prot->state, PROT_SYSCTL_INSTALLED)) {
			unregister_sysctl_table(prot->sysctl_table_header);
			PR_DEBUG("removed %s sysctl handler", prot->name);
		}

#ifdef CONFIG_PROC_FS
		/* unregister procfs operations */
		if (PROTOCOL_STATE(prot->state, PROT_PROCFS_INSTALLED)) {
			proc_net_remove(INIT_NET_ARG prot->name);
			PR_DEBUG("removed %s procfs operations", prot->name);
		}
#endif
		/* unregister packet receiving kernel hook */
		if (PROTOCOL_STATE(prot->state, PROT_RECVHOOK_INSTALLED)) {
			dev_remove_pack(prot->receive_hook);
			PR_DEBUG("removed %s receiving kernel hook", prot->name);
		}
		/* unregister netdevice notifier */
		if (PROTOCOL_STATE(prot->state, PROT_NOTIFIER_INSTALLED)) {
			err = unregister_netdevice_notifier(prot->netdev_notifier);
			if (err < 0)
				PR_ERR("could not remove %s network device notifier", prot->name);
			PR_DEBUG("removed %s network device notifier", prot->name);
		}
		/* call protocol destructor */
		if (PROTOCOL_STATE(prot->state, PROT_INITIALIZED) && prot->destroy) {
			prot->destroy();
			PR_DEBUG("destroyed protocol %s", prot->name);
		}
	}


	/* unregister protocol family */
	if (pf_installed) {
		sock_unregister(PF_ENET);
		PR_DEBUG("removed protocol family %d", PF_ENET);
	}
}


/*
 * Initialization function called upon module loading.
 */
static int __init enet_init(void)
{
	int i;
	int err = 0;

	/* register protocol family operations for the new socket type. */
	err = sock_register(&enet_family_ops);
	if (err < 0) {
		PR_ERR("could not register protocol family %d", PF_ENET);
		return err;
	}
	pf_installed = 1;
	PR_DEBUG("registered protocol family %d", PF_ENET);

	/* register all protocols with the kernel */
	for (i = 0; i < ENET_PROTOCOL_COUNT; i++) {
		struct enet_protocol *prot = &enet_protocols[i];
		/* call protocol initialization function */
		if (prot->init) {
			err = prot->init();
			if (err < 0) {
				PR_ERR("could not initialize protocol %s", prot->name);
				goto out_cleanup;
			}
			PR_DEBUG("initialized protocol %s", prot->name);
		}
		++prot->state;

		/* register netdevice notifier, which is called by the
		 * kernel upon netdevice state changes
		 */
		if (prot->netdev_notifier) {
			err = register_netdevice_notifier(prot->netdev_notifier);
			if (err < 0) {
				PR_ERR("could not install %s network device notifier", prot->name);
				goto out_cleanup;
			}
			PR_DEBUG("installed %s network device notifier", prot->name);
			++prot->state;
		}

		/* install packet receiving kernel hook */
		dev_add_pack(prot->receive_hook);
		PR_DEBUG("installed %s receiving kernel hook for packet type %#x",
		         prot->name,
		         ntohs(prot->receive_hook->type));
		++prot->state;

#ifdef CONFIG_PROC_FS
		/* register file operations for the /proc filesystem */
		if (prot->proc_fops) {
			proc_net_fops_create(INIT_NET_ARG prot->name, 0, prot->proc_fops);
			PR_DEBUG("installed %s procfs operations", prot->name);
			++prot->state;
		}
#endif

		if (prot->sysctl_table) {
			if (!(prot->sysctl_table_header =
			                        register_sysctl_table(prot->sysctl_table))) {
				PR_ERR("Could not install %s sysctl handler.", prot->name);
				goto out_cleanup;
			}
			PR_DEBUG("installed %s sysctl handler.", prot->name);
			++prot->state;
		}


	}

	return 0;

out_cleanup:
	module_cleanup();

	return err;
}



/*
 * Shutdown function called upon module unloading.
 */
static void __exit enet_exit(void) {
	module_cleanup();
}


module_init(enet_init);
module_exit(enet_exit);

MODULE_AUTHOR("Mirko Reinhardt");
MODULE_DESCRIPTION("Implements a raw ethernet protocol family with unreliable and reliable datagram transport.");
MODULE_LICENSE("GPL");
MODULE_ALIAS_NETPROTO(PF_ENET);
