/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

/*
 * ENET  An implementation of an Ethernet protocol suite for the LINUX
 *  operating system.  ENET is implemented using the  BSD Socket
 *  interface as the means of communication with the user level.
 *
 *  ESP - ethernet streaming protocol
 *  Implements a reliable and sequenced means to send and receive
 *   data
 *
 * Version: $Id: enet_stream.c,v 0.1 2005/xx/xx 10:00:00 rmir Exp $
 *
 * Authors: Mirko Reinhardt, <rmir@hrz.tu-chemnitz.de>
 *   Matthias Treydte, <waldheinz@gmail.com>
 *
 * Fixes:
 */

#include <linux/net.h>  /* struct socket, struct proto_ops */
#include <net/sock.h>  /* struct sock, basic socket operations */
#include <linux/netdevice.h> /* struct net_device, basic netdevice ops */
#include <linux/skbuff.h> /* socket buffer struct and ops */
#include <linux/timer.h> /* struct timer_list and timer ops */
#include <linux/list.h>  /* list operations */
#include <asm/ioctls.h>  /* IOCTL constants */

#include "enet_module.h" /* project-global include */
#include "enet_socket.h" /* ENET socket global ops */
#include "enet_stream.h"
#include "enet_stream_input.h"
#include "enet_stream_output.h"
#include "enet_stream_cong.h"
#include "enet_compat.h"


/*************************************************************************
 * Internal data structures
 ************************************************************************/


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11)

/* esp protocol descriptor */
struct proto esp_proto = {
	.name     = "ESP",
	.owner    = THIS_MODULE,
	.obj_size = sizeof(struct esp_sock),
};

#else /* pre-2.6.12 kernels */


/* esp socket slab-cache */
static kmem_cache_t *esp_sk_cachep = NULL;


/* create the socket slab cache */
int esp_prot_init_slab(void)
{
	int rc = 0;

	/* create socket slab cache */
	esp_sk_cachep =
	        kmem_cache_create("esp_sock",
	                          sizeof(struct esp_sock), 0, SLAB_HWCACHE_ALIGN, NULL, NULL);

	if (!esp_sk_cachep) {
		PR_ERR("could not create esp socket slab cache");
		rc = -ENOMEM;
	}

	return rc;
}

/**
 * Destroy socket slab cache
 */
void esp_prot_destroy_slab(void)
{
	if (esp_sk_cachep)
		kmem_cache_destroy(esp_sk_cachep);
}

/**
 * ESP socket allocation function
 */
struct sock *esp_sk_alloc(int priority)
{
	return sk_alloc(PF_ENET, priority, sizeof(struct esp_sock),
	                esp_sk_cachep);
}

#define ESP_SK_USES_WRITE_QUEUE(__sk) \
 (__sk)->sk_use_write_queue = 1

#endif /* pre-2.6.12 kernels */

/** counter of ESP sockets alive */
atomic_t esp_skcount;

/** bound sockets hashtable */
struct hlist_head esp_bhash[ESP_BHTABLE_SIZE];
/** listening sockets hashtable */
struct hlist_head esp_lhash[ESP_LHTABLE_SIZE];
/** established connection sockets hashtable */
struct hlist_head esp_ehash[ESP_EHTABLE_SIZE];

/* port-to-socket hashtable r/w-lock */
DEFINE_RWLOCK(esp_bhash_lock);
DEFINE_RWLOCK(esp_lhash_lock);
DEFINE_RWLOCK(esp_ehash_lock);

/* port range to automatically choose ports from and port rover */
int esp_local_port_range[2] = { 1, 65535 };
int esp_port_rover = 0;


/*
 * Initialize some values of an esp socket going to connect
 */
static inline void esp_sk_init_connect(struct sock *sk, unsigned char *raddr,
                                       en_port_t rport, unsigned short recv_seq,
                                       unsigned char passive)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* set peer address and port */
	memcpy(esp_sk->raddr, raddr, ETH_ALEN);
	esp_sk->rport = rport;

	/* initialize sequence numbers */
	esp_sk->recv_seq = recv_seq;
	esp_sk->send_seq = net_random();
	esp_sk->una_seq  = esp_sk->send_seq;
	esp_sk->ack_seq  = esp_sk->recv_seq - passive;
	esp_sk->copy_seq = recv_seq;
	esp_sk->copy_off = 0;

	esp_sk->rcv_mem_pressure = 0;
	esp_sk->trans_announced = 0;
	esp_sk->announce_seq = 0;
	esp_sk->recv_announced = 0;
	esp_sk->need_rtx = 0;

	/* initialize rtt values */
	esp_sk->srtt     = 0;
	esp_sk->rto      = ESP_RTO_INIT;
}

/*
 * Insert the given socket into the bound hashtable
 * The bound hashtable uses ports in host byte order.
 * This function assumes that we already hold a lock on the bound hash table!
 */
static inline void esp_sk_hash_bound(struct sock *sk, struct net_device *dev,
                                     unsigned short num)
{

	struct esp_opt *esp_sk = ESP_SK(sk);
	struct hlist_head *h = &esp_bhash[ESP_BHASH_FN(num)];

	PR_DEBUG("hashing esp socket %p for port %d into bucket %d",
	         sk, num, ESP_BHASH_FN(num));

	sk_add_bind_node(sk, h);
	esp_sk->port = num;
	esp_sk->lport = htons(num);

	if (dev) {
		esp_sk->ifindex = dev->ifindex;
		memcpy(&esp_sk->laddr, &dev->dev_addr, ETH_ALEN);
	} else {
		esp_sk->ifindex = 0;
	}

	esp_sk->bound = 1;
}

/*
 * Insert the given socket into the listen hashtable
 * The listening hashtable uses ports in network byte order
 * to avoid the byte order conversion during socket lookups.
 * This function grabs a write lock on the listening hash table.
 */
static inline void esp_sk_hash_listen(struct sock *sk)
{
	printk("hashing %p into listen\n", sk);
	write_lock_bh(&esp_lhash_lock);
	BUG_ON(sk_hashed(sk));

	if (sk_unhashed(sk)) {
		struct esp_opt *esp_sk = ESP_SK(sk);
		u16 hash = esp_lhash_fn(esp_sk->lport);
		__sk_add_node(sk, &esp_lhash[hash]);
		PR_DEBUG("hashed listening esp socket %p into bucket %d",
		         sk, hash);
	} else
		PR_ERR("esp socket %p already hashed", sk);

	write_unlock_bh(&esp_lhash_lock);
}

/*
 * Insert the given socket into the established hashtable
 * The established hashtable uses remote and local addresses and ports in
 * network byte order to avoid the byte order conversion during socket lookups.
 * The hash function performs several XORs to provide a good dispersion.
 * This function grabs a write lock on the established hash table.
 */
static inline void esp_sk_hash_established(struct sock *sk)
{
	printk("hashing socket %p into established table\n", sk);
	write_lock_bh(&esp_ehash_lock);

	if (sk_unhashed(sk)) {
		struct esp_opt *esp_sk = ESP_SK(sk);
		u16 hash = esp_ehash_fn(esp_sk->laddr, esp_sk->raddr,
		                        esp_sk->lport, esp_sk->rport);
		__sk_add_node(sk, &esp_ehash[hash]);
		PR_DEBUG("hashed connecting / connected esp socket %p into bucket %d",
		         sk, hash);
	} else
		PR_ERR("esp socket %p already hashed", sk);

	write_unlock_bh(&esp_ehash_lock);
}

/*
 * Determine whether the given port is free to use
 * or (if rport == 0) the next free port to use
 */
static int esp_get_port(struct sock *sk, struct net_device *dev,
                        unsigned short num)
{
	struct hlist_node *node;
	struct sock *sk2;

	PR_DEBUG("Looking for a free port, supplied port num: %u", num);

	write_lock_bh(&esp_bhash_lock);

	if (num == 0) {
		/* user suppled 0 for port - determine a good free one */
		int best_size_so_far, best, result, i;

		if (esp_port_rover > esp_local_port_range[1] ||
		                esp_port_rover < esp_local_port_range[0])
			esp_port_rover = esp_local_port_range[0];
		best_size_so_far = 32767;
		best = result = esp_port_rover;
		/* look for the next empty or the shortest queue in the hashtable */
		for (i = 0; i < ESP_BHTABLE_SIZE; i++, result++) {
			struct hlist_head *list;
			int size;

			list = &esp_bhash[ESP_BHASH_FN(result)];
			if (hlist_empty(list)) {
				if (result > esp_local_port_range[1])
					result = esp_local_port_range[0] +
					         ESP_BHASH_FN(result - esp_local_port_range[0]);
				goto gotit;
			}
			size = 0;
			sk_for_each_bound(sk2, node, list)
			if (++size >= best_size_so_far)
				goto next;
			best_size_so_far = size;
			best = result;
next:
			;
		}
		/* determine the next free port in the found hashtable queue */
		result = best;
		for (i = 0; i < (1 << 16) / ESP_BHTABLE_SIZE; i++, result += ESP_BHTABLE_SIZE) {
			if (result > esp_local_port_range[1])
				result = esp_local_port_range[0]
				         + ESP_BHASH_FN(result - esp_local_port_range[0]);
			if (!esp_port_inuse(result))
				break;
		}
		if (i >= (1 << 16) / ESP_BHTABLE_SIZE)
			goto fail;
gotit:
		num = result;
		/* automatically choose consecutive ports */
		esp_port_rover = num + 1;
	} else {
		/* user supplied a port - check if it can be used */
		sk_for_each_bound(sk2, node, &esp_bhash[ESP_BHASH_FN(num)]) {
			struct esp_opt *esp_sk2 = ESP_SK(sk2);
			PR_DEBUG("esp_get_port: checking other socket for same port: %p", esp_sk2);
			if (esp_sk2->port == num && sk2 != sk &&
			                (!esp_sk2->ifindex || !dev ||
			                 esp_sk2->ifindex == dev->ifindex))
				goto fail;
		}
	}

	/* insert the socket into the hashtable */
	if (sk_unbound(sk)) {
		esp_sk_hash_bound(sk, dev, num);
	}

	write_unlock_bh(&esp_bhash_lock);
	PR_DEBUG("<- esp_get_port (succeeded)");
	return 0;

fail:
	write_unlock_bh(&esp_bhash_lock);
	PR_DEBUG("<- esp_get_port (failed)");
	return 1;
}

/*
 * Lookup a socket in the established hashtable
 */
static inline struct sock *esp_sk_lookup_established(unsigned char *src_mac,
			                unsigned char *dst_mac,
			                unsigned short src_port,
			                unsigned short dst_port,
			                int dst_if) {

	struct sock *sk = NULL;
	struct hlist_node *node;
	u16 hash = esp_ehash_fn(dst_mac, src_mac, dst_port, src_port);

	read_lock(&esp_ehash_lock);

	PR_DEBUG("searching esp socket in established bucket %d", hash);

	sk_for_each(sk, node, &esp_ehash[hash]) {
		struct esp_opt *esp_sk = ESP_SK(sk);
		if (!ETH_ADDR_CMP(dst_mac, esp_sk->laddr) &&
		                !ETH_ADDR_CMP(src_mac, esp_sk->raddr) &&
		                dst_port == esp_sk->lport &&
		                src_port == esp_sk->rport &&
		                (esp_sk->ifindex == 0 || esp_sk->ifindex == dst_if)) {
			sock_hold(sk);
			read_unlock(&esp_ehash_lock);
			return sk;
		}
	}

	read_unlock(&esp_ehash_lock);

	return NULL;
}

/*
 * Lookup a socket in the listen hashtable
 * The destination port is expected in network byte order
 */
static inline struct sock *esp_sk_lookup_listen(unsigned short dst_port,
			                int dst_if) {
	struct sock *sk = NULL;
	struct hlist_node *node;
	u16 hash = esp_lhash_fn(dst_port);

	read_lock(&esp_lhash_lock);

	PR_DEBUG("searching esp socket in listen bucket %d", hash);

	sk_for_each(sk, node, &esp_lhash[hash]) {
		struct esp_opt *esp_sk = ESP_SK(sk);
		if (esp_sk->lport == dst_port &&
		                (esp_sk->ifindex == 0 || esp_sk->ifindex == dst_if)) {
			sock_hold(sk);
			read_unlock(&esp_lhash_lock);
			return sk;
		}
	}

	read_unlock(&esp_lhash_lock);

	return NULL;
}

/*
 * Lookup a socket in the established/listen hashtables
 * All values are expected in network byte order (except dst_if ;)
 */
static inline struct sock *esp_sk_lookup(void *src_mac,
			                void *dst_mac,
			                unsigned short src_port,
			                unsigned short dst_port,
			                int dst_if) {
	struct sock *sk;

	/* first try to find a connected socket and if that fails try
	 * to find a listening one for the provided addresses and ports */
	sk = esp_sk_lookup_established(src_mac, dst_mac, src_port, dst_port, dst_if);
	return sk ? : esp_sk_lookup_listen(dst_port, dst_if);
}

/*
 * Enqueue a child socket to the syn or acc queue of its
 * listening parent socket
 */
void esp_openreq_enqueue(struct sock *parent, struct sock *child)
{
	struct esp_opt *parent_sk = ESP_SK(parent);
	struct esp_opt *child_sk  = ESP_SK(child);

	child_sk->parent = parent;

	switch (child->sk_state) {
	case ESP_SYN_RECV:
		list_add_tail(&child_sk->syn_queue, &parent_sk->syn_queue);
		esp_synq_added(parent);
		PR_DEBUG("added child %p to syn queue of esp sock %p "
		         "(synq %d accq %d)",
		         child, child_sk->parent, parent_sk->synq_len,
		         parent->sk_ack_backlog);
		break;

	case ESP_ESTABLISHED:
		list_add_tail(&child_sk->acc_queue, &parent_sk->acc_queue);
		sk_acceptq_added(parent);
		PR_DEBUG("added child %p to acc queue of esp sock %p "
		         "(synq %d accq %d)",
		         child, child_sk->parent, parent_sk->synq_len,
		         parent->sk_ack_backlog);
		break;
	}
}

/*
 * Dequeue a child socket from the syn or acc queue of its
 * listening parent socket
 */
void esp_openreq_dequeue(struct sock *child)
{
	struct esp_opt *child_sk = ESP_SK(child);
	struct sock *parent = child_sk->parent;

	if (parent && !list_empty(&child_sk->syn_queue)) {
		list_del_init(&child_sk->syn_queue);
		child_sk->parent = NULL;
		esp_synq_removed(parent);

		PR_DEBUG("removed child %p from syn queue of esp sock %p "
		         "(synq %d accq %d)",
		         child, parent, ESP_SK(parent)->synq_len,
		         parent->sk_ack_backlog);
	}

	if (parent && !list_empty(&child_sk->acc_queue)) {
		list_del_init(&child_sk->acc_queue);
		child_sk->parent = NULL;
		sk_acceptq_removed(parent);

		PR_DEBUG("removed child %p from acc queue of esp sock %p (synq %d accq %d)",
		         child, parent, ESP_SK(parent)->synq_len,
		         parent->sk_ack_backlog);
	}
}

/*
 * Spawn a child socket due to a connection request on a listening socket
 */
struct sock *esp_sk_spawn_child(struct sock *parent,
			                                struct sk_buff *skb) {
	struct esp_opt *parent_sk = ESP_SK(parent);
	struct sock *child;
	struct esp_opt *child_sk;
	struct ethhdr *eth_hdr = (struct ethhdr *) skb_mac_header(skb);
	struct esp_header *esp_hdr = (struct esp_header *) skb_network_header(skb);

	child = esp_sk_alloc(INIT_NET_ARG GFP_ATOMIC);
	if (!child) {
		PR_ERR("Could not allocate child socket for esp socket %p", parent);
		goto out;
	}

	/* init kernel sock structure (no associated socket structure yet) */
	esp_sk_init(NULL, child);

	child->sk_protocol = parent->sk_protocol;
	atomic_inc(&esp_skcount);

	child_sk = ESP_SK(child);

	PR_DEBUG("esp child socket %p created, %d are alive", child,
	         atomic_read(&esp_skcount));

	esp_set_state(child, ESP_SYN_RECV);

	/* enqueue the child socket into the parent socket's syn queue */
	esp_openreq_enqueue(parent, child);

	/* hash the child into the bound and established hashtables,
	 * initialize the socket's connection values */
	write_lock_bh(&esp_bhash_lock);
	esp_sk_hash_bound(child, SKB_INPUT_DEV(skb), parent_sk->port);
	write_unlock_bh(&esp_bhash_lock);

	esp_sk_init_connect(child, eth_hdr->h_source, esp_hdr->src_port,
	                    ESP_SKB_CB(skb)->pkt_seq + 1, 1);
	esp_sk_hash_established(child);

out:
	return child;
}


/*
 * Destroy a child socket from the syn or acc queue of a listening socket
 */
static inline void esp_sk_destroy_child(struct sock *child)
{
	sock_hold(child);
	bh_lock_sock(child);

	esp_set_state(child, ESP_CLOSE);
	sock_orphan(child);
	esp_sk_destroy(child);

	bh_unlock_sock(child);
	sock_put(child);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
static inline void enet_write_queue_purge(struct sock *sk)
{
	struct sk_buff *skb;

	while ((skb = __skb_dequeue(&sk->sk_write_queue)) != NULL) {
		sk_wmem_free_skb(sk, skb);
	}
	sk_mem_reclaim(sk);
}
#endif

void esp_sk_destroy(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sock *child;
	struct esp_opt *child_sk, *tmp;
	struct sk_buff *skb;
	int data_unread = 0;

	printk("destroying esp socket %p\n", sk);

	WARN_ON(!sk_unhashed(sk));

	/* stop retransmission and RRQ timers */
	sk_stop_timer(sk, &esp_sk->rtx_timer);
	sk_stop_timer(sk, &esp_sk->rrq_timer);

	/* clear accept queues */
	list_for_each_entry_safe(child_sk, tmp, &esp_sk->syn_queue, syn_queue) {
		child = SK(child_sk);
		PR_DEBUG("destroying syn child %p of esp socket %p", child, sk);
		esp_sk_destroy_child(child);
	}

	/* destroy waiting children */
	list_for_each_entry_safe(child_sk, tmp, &esp_sk->acc_queue, acc_queue) {
		child = SK(child_sk);
		PR_DEBUG("destroying acc child %p of esp socket %p", child, sk);
		esp_sk_destroy_child(child);
	}

	/* clear out-of-sequence packet queue */
#ifdef ESP_USE_OOS_QUEUE
	__skb_queue_purge(&esp_sk->oos_packets);
#endif

	/* clear packet retransmission queue */
	if (!skb_queue_empty(&sk->sk_write_queue)) {
		PR_DEBUG("flushing esp socket %p retransmission queue", sk);
		enet_write_queue_purge(sk);
	}

	/* maybe something was enqueued but not eaten by userspace, purge it */

	/* purge receive queue */
	while ((skb = __skb_dequeue(&sk->sk_receive_queue)) != NULL) {
		data_unread += skb->len;
		__kfree_skb(skb);
	}

	if (data_unread) {
		/* we dropped some data unseen by the user -
		 * TODO: we should send a RST */
		PR_DEBUG("purged %d bytes of unread data from esp socket %p (rmem now: %d)",
		         data_unread, sk, atomic_read(&sk->sk_rmem_alloc));
	}


	/* cleanup ack_queue */
	esp_purge_ack(sk);

	/* cancel this transmission */
	if (esp_sk->recv_announced && ! esp_sk->rcv_mem_pressure) {
		atomic_dec(&esp_recv_count);
		esp_sk->recv_announced = 0;
	}
	sock_put(sk);
}

/*
 * Convenience function for SKB queue retransmission
 */
static inline void esp_write_error(struct sock *sk, int err)
{
	SK_ERR_REPORT(sk, err);
	esp_set_state(sk, ESP_CLOSE);
	if (!sock_flag(sk, SOCK_DEAD)) {
		sk->sk_state_change(sk);
	} else {
		esp_sk_destroy(sk);
	}
}


/*
 * resets the retransmission timer interval and (re)starts the timer
 * { performs sock_hold(sk) }
 */
void esp_rtx_timer_reset(struct sock *sk, unsigned long when)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* function performs  sock_hold(sk) */
	sk_reset_timer(sk, &esp_sk->rtx_timer, ESP_TIMESTAMP + when);

	PR_DEBUG("reset rtx timeout to %lu on esp socket %p", when, sk);
}


/* resets the retransmission timer interval for the
 * FIN_WAIT timeout and (re)starts the timer, so the peer
 * can gracefully close his side of the connection, but can't
 * cause our socket to linger forever waiting for the FIN
 * { performs sock_hold(sk) }
 */
void esp_rtx_timer_finwait(struct sock *sk)
{
	sk_reset_timer(sk, &ESP_SK(sk)->rtx_timer,
	               ESP_TIMESTAMP + ESP_FINWAIT2_TO);

	PR_DEBUG("reset fin-wait timeout to %lu on esp socket %p",
	         ESP_TIMESTAMP + ESP_FINWAIT2_TO, sk);
}


/*
 * Retransmission timer callback function.
 */
void esp_rtx_timer(unsigned long data)
{
	struct sock *sk = (struct sock *) data;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sk_buff *skb;
	int err;

	bh_lock_sock(sk);

	/* we don't do retransmissions in these cases! */
	if ((sk->sk_state == ESP_CLOSE) || (sk->sk_state == ESP_CLOSE_WAIT))
		goto out;

	if (sock_owned_by_user(sk)) {
		/* try again later if the socket is in the user's hands */
		esp_rtx_timer_reset(sk, ESP_RESOURCE_TO);
		goto out;
	}

	/* check for the FINWAIT timeout */
	if (sk->sk_state == ESP_FIN_WAIT2) {
		/* restart the timer if the socket has not been closed yet */
		if (!sock_flag(sk, SOCK_DEAD)) {
			esp_rtx_timer_finwait(sk);
			goto out;
		}
		PR_DEBUG("fin wait timeout of esp socket %p expired - "
		         "closing socket", sk);
		esp_set_state(sk, ESP_CLOSE);


		esp_sk_destroy(sk);
		goto out;
	}

	/* check for final socket write timeout */
	if ((ESP_TIMEOUT_CONN) && (esp_sk->rto >= ESP_MAX_RTO)) {
		PR_DEBUG("maximum rtx timeout exceeded on esp socket %p "
		         "(rto %lu ts %lu)", sk, esp_sk->rto, ESP_TIMESTAMP);
		esp_write_error(sk, -ETIMEDOUT);
		goto out;
	}

	/* check if there is at least one packet to retransmit */
	skb = skb_peek(&sk->sk_write_queue);
	if (!skb) {
		PR_DEBUG("unnecessary rtx timeout on esp socket %p (ts %lu)",
		         sk, ESP_TIMESTAMP);
		goto out;
	}

	/* check for final write timeout
	 * the first packet in the queue is the oldest and most retransmitted
	 * one in the queue so we only have to check it */
	if ((ESP_MAX_RETRANSMITS) &&
	                (ESP_SKB_CB(skb)->xmits > ESP_MAX_RETRANSMITS)) {
		PR_DEBUG("maximum rtx count exceeded on esp socket %p "
		         "(seq %u xmits %d rto %lu ts %lu)",
		         sk, ESP_SKB_CB(skb)->pkt_seq, ESP_SKB_CB(skb)->xmits,
		         esp_sk->rto, ESP_TIMESTAMP);
		esp_write_error(sk, (err = -ETIMEDOUT));
		goto out;
	}

	/* update the socket's timeout value */
	esp_double_rto(sk);

	/* do the actual retransmission, stop timer
	 * if there was nothing to do
	 */

	if (/* got data? */
	        (sk->sk_write_queue.next != (struct sk_buff *)&sk->sk_write_queue) &&
	        /* got _old_ data? */
	        (sk->sk_send_head && (sk->sk_send_head != sk->sk_write_queue.next))) {

		skb = sk->sk_write_queue.next;

		do {
			/* we may push frames up to (excluding) the send head */
			if ((skb == (struct sk_buff *)&sk->sk_write_queue) ||
			                (skb == sk->sk_send_head))
				break;

			printk("rtx-timer pkt_seq: %d, flags: "ESP_FLAGS_FORMAT"send_seq: %d sock: %p \n",
			       ESP_SKB_CB(skb)->pkt_seq, ESP_FLAGS_DEBUG(ESP_HDR(skb)->flags),
			       esp_sk->send_seq, sk);

			esp_skb_xmit(sk, skb_clone(skb, GFP_ATOMIC));
			skb = skb->next;

		} while (1);

	} else {
		/* we didn't have anything to retransmit */
		goto out;
	}

	/* restart timer */
	esp_rtx_timer_reset(sk, ESP_LONG_TIMEOUT);

out:
	bh_unlock_sock(sk);
	sock_put(sk);
}

/*
 * Shutdown a socket
 */
static inline int esp_sk_shutdown(struct sock *sk)
{
	int err = 0;
	PR_DEBUG("-> esp_sk_shutdown (current state: %s)", esp_statenames[sk->sk_state]);

	/* send a FIN in synchronized states */
	if (SK_STATE(sk, ESPF_SYN_RECV | ESPF_ESTABLISHED | ESPF_CLOSE_WAIT)) {
		PR_DEBUG("Sending final FIN on socket %p.", sk);
		err = esp_send_fin(sk, sk->sk_allocation);
	}

	/* set the new socket state */
	switch (sk->sk_state) {
	case ESP_ESTABLISHED:
	case ESP_SYN_RECV:
		esp_set_state(sk, ESP_FIN_WAIT1);
		break;
	case ESP_CLOSE_WAIT:
		esp_set_state(sk, ESP_LAST_ACK);
		break;
	case ESP_SYN_SENT:
	case ESP_TIME_WAIT:
	case ESP_LISTEN:
		esp_set_state(sk, ESP_CLOSE);
		break;
	case ESP_CLOSE:
		esp_set_state(sk, ESP_CLOSE);
	}

	PR_DEBUG("<- esp_sk_shutdown");
	return err;
}

/*
 * RRQ timer callback function
 */
static void esp_rrq_timer(unsigned long data)
{
	struct sock *sk = (struct sock *) data;
	struct esp_opt *esp_sk = ESP_SK(sk);
	int err;

	bh_lock_sock(sk);

	if ((sk->sk_state == ESP_CLOSE) ||
	                (sk->sk_state == ESP_CLOSE_WAIT) ||
	                (!esp_sk->recv_announced))
		goto out;

	/* if we still have data pending there's no point
	 * requesting more.
	 */
	if (sk->sk_backlog.tail)
		goto out;

	if (sock_owned_by_user(sk)) {
		/* try again later */
		ESP_RESET_RRQ_TIMER(sk, ESP_RESOURCE_TO);
		printk("rrq while in users hand; backlock empty: %d\n",
		       sk->sk_backlog.tail ? 1 : 0);
		goto out;
	}

	printk("RRQ timeout on esp socket %p seq %d\n", sk, ESP_SK(sk)->recv_seq);

	err = esp_send_rrq(sk);

	/* try again later (but yet very soon) on error */
	if (err) {
		printk("error while sending rrq in rrq timer?!\n");
		ESP_RESET_RRQ_TIMER(sk, ESP_RESOURCE_TO);
	} else {
		ESP_RESET_RRQ_TIMER(sk, esp_round_trip_time);
	}

out:
	bh_unlock_sock(sk);
	sock_put(sk);
}

/*
 * connect() waiting function for blocking sockets
 */
static long esp_connect_wait(struct sock *sk, long timeo)
{
	DEFINE_WAIT(wait); /* wait queue for blocking sockets */

	prepare_to_wait(sk->sk_sleep, &wait, TASK_INTERRUPTIBLE);

	/* socket state MUST be changed from ESP_SYN_* when sk->sk_err is set
	 * or connect() won't return...
	 */
	while (SK_STATE(sk, ESPF_SYN_SENT | ESPF_SYN_RECV)) {
		release_sock(sk);
		timeo = schedule_timeout(timeo);
		lock_sock(sk);
		if (signal_pending(current) || !timeo)
			break;
		prepare_to_wait(sk->sk_sleep, &wait, TASK_INTERRUPTIBLE);
	}

	finish_wait(sk->sk_sleep, &wait);
	return timeo;
}

/*
 * internal connect function
 */
static int esp_connect(struct sock *sk, struct sockaddr *uaddr,
                       int addr_len)
{
	struct net_device *dev;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sockaddr_en *sen = (struct sockaddr_en *) uaddr;
	int err = 0;

	PR_DEBUG("-> esp_connect");

	dev = dev_get_by_index(INIT_NET_ARG sen->sen_ifindex);
	if (!dev) {
		PR_DEBUG("<- esp_connect (ENODEV)");
		return -ENODEV;
	}

	/* report an error if the device is down */
	if (!(dev->flags & IFF_UP)) {
		dev_put(dev);
		PR_DEBUG("<- esp_connect (ENETDOWN)");
		return -ENETDOWN;
	}

	PR_DEBUG("connecting esp socket %p port %d if %s -> "NETH_FORMAT" port %d",
	         sk, esp_sk->port, dev->name, NETH_HEX(sen->sen_addr),
	         ntohs(sen->sen_port));

	if (esp_get_port(sk, dev, esp_sk->port)) {
		err = -EADDRNOTAVAIL;
		goto out;
	}

	sk->sk_err = 0;

	/* initialize the socket's connection values */
	esp_sk_init_connect(sk, sen->sen_addr, sen->sen_port, 0, 0);

	/* preliminarily insert the socket into the established hashtable so
	 * the peer's SYN-ACK can be properly received */
	esp_sk_hash_established(sk);

	esp_send_syn(sk, dev, sk->sk_allocation);

	esp_set_state(sk, ESP_SYN_SENT);
	dev_put(dev);
	PR_DEBUG("<- esp_connect");
	return err;

out:
	dev_put(dev);
	memset(&esp_sk->raddr, 0, ETH_ALEN);
	esp_sk->rport = 0;
	esp_set_state(sk, ESP_CLOSE);
	PR_DEBUG("<- esp_connect (error)");
	return err;
}



/*
 * BIND handler
 */
static int sock_esp_bind(struct socket *sock, struct sockaddr *uaddr,
                         int addr_len)
{
	struct sockaddr_en *sen = (struct sockaddr_en*)uaddr;
	struct sock *sk = sock->sk;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct net_device *dev = NULL;
	unsigned short port;
	int err = 0;

	PR_DEBUG("-> sock_esp_bind");

	/* check legality of given address */
	if (addr_len < sizeof(*sen))
		return -EINVAL;

	if (sen->sen_family != AF_ENET)
		return -EAFNOSUPPORT;

	/* undo the byte shuffling for ports hashtable */
	port = ntohs(sen->sen_port);

	PR_DEBUG("bind esp socket %p to interface %d port %u", sk, sen->sen_ifindex, port);

	/* bind to specific interface or all interfaces using ifindex == 0 */
	if (sen->sen_ifindex) {
		dev = dev_get_by_index(INIT_NET_ARG sen->sen_ifindex);
		if (!dev) {
			err = -ENODEV;
			goto out;
		}
	}

	/* report an error if the device is down */
	if (dev && !(dev->flags & IFF_UP)) {
		err = -ENETDOWN;
		SK_ERR_REPORT(sk, err);
		goto out_put;
	}

	lock_sock(sk);

	/* report an error if already bound */
	if (esp_sk->bound) {
		err = -EINVAL;
		goto out_release;
	}

	/* Check the port and bind */
	if (esp_get_port(sk, dev, port)) {
		err = -EADDRINUSE;
		goto out_release;
	}

	PR_DEBUG("bound esp socket %p to interface %s port %u",
	         sk, dev ? dev->name : "0", esp_sk->port);
	PR_DEBUG("<- sock_esp_bind");

out_release:
	release_sock(sk);
out_put:
	if (dev)
		dev_put(dev);
out:
	return err;
}

/*
 * LISTEN handler
 */
static int sock_esp_listen(struct socket *sock, int backlog)
{
	struct sock *sk = sock->sk;
	int err = 0;

	PR_DEBUG("esp socket %p entering listening mode", sk);

	lock_sock(sk);

	if (sock->state != SS_UNCONNECTED) {
		PR_ERR("esp socket %p not unconnected (state %d)", sk,
		       sock->state);
		err = -EINVAL;
		goto out;
	}

	if (!SK_STATE(sk, ESPF_CLOSE | ESPF_LISTEN)) {
		PR_ERR("esp socket %p in wrong state %d", sk, sk->sk_state);
		err = -EINVAL;
		goto out;
	}

	if (sk_unbound(sk) && !esp_get_port(sk, NULL, 0)) {
		err = -EAGAIN;
		goto out;
	}

	/* if the socket is already in listening state only the backlog
	 * can be adjusted */
	if (sk->sk_state != ESP_LISTEN) {
		esp_sk_hash_listen(sk);
		esp_set_state(sk, ESP_LISTEN);
	}

	sk->sk_max_ack_backlog = backlog;
	err = 0;

out:
	release_sock(sk);
	return err;
}


/*
 * ACCEPT handler
 */
static int sock_esp_accept(struct socket *sock, struct socket *newsock,
                           int flags)
{
	struct sock *sk = sock->sk;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sock *newsk = NULL;
	DEFINE_WAIT(wait);
	int err = 0;

	PR_DEBUG("accepting a connection on esp socket %p", sk);

	lock_sock(sk);

	if (sk->sk_state != ESP_LISTEN) {
		err = -EINVAL;
		goto out_release;
	}

	if (list_empty(&esp_sk->acc_queue)) {
		long timeo = sock_rcvtimeo(sk, flags & O_NONBLOCK);

		/* wait for incoming connections on blocking sockts */
		for (;;) {
			prepare_to_wait_exclusive(sk->sk_sleep, &wait,
			                          TASK_INTERRUPTIBLE);
			err = sock_error(sk) ? : -EINVAL;
			if (sk->sk_state != ESP_LISTEN)
				break;
			err = sock_intr_errno(timeo);
			if (signal_pending(current))
				break;
			err = -EAGAIN;
			if (!timeo)
				break;
			err = 0;
			/* wait for an acceptible connection */
			if (sk_wait_event(sk, &timeo, !list_empty(&esp_sk->acc_queue)))
				break;
		}

		finish_wait(sk->sk_sleep, &wait);

		if (err) {
			PR_DEBUG("accept failed on esp socket %p with err %d",
			         sk, err);
			goto out_release;
		}
	}

	newsk = SK(list_entry(esp_sk->acc_queue.next, struct esp_opt, acc_queue));
	esp_openreq_dequeue(newsk);

	PR_DEBUG("accepted connection on child %p of esp socket %p", newsk, sk);

out_release:
	release_sock(sk);

	if (!newsk)
		goto out;

	lock_sock(newsk);

	WARN_ON(newsk->sk_state == ESP_SYN_RECV);

	sock_graft(newsk, newsock);
	newsock->state = SS_CONNECTED;
	err = 0;

	release_sock(newsk);

out:
	return err;
}


/*
 * CONNECT handler
 */
static int sock_esp_connect(struct socket *sock, struct sockaddr *uaddr,
                            int addr_len, int flags)
{
	struct sock *sk = sock->sk;
	struct sockaddr_en *sen = (struct sockaddr_en *) uaddr;
	int err = 0;
	long timeo;
	PR_DEBUG("-> sock_esp_connect");

	if (addr_len < sizeof(struct sockaddr_en)) {
		PR_DEBUG("<- sock_esp_connect (EINVAL)");
		return -EINVAL;
	}

	if (sen->sen_family != AF_ENET) {
		PR_DEBUG("<- sock_esp_connect (EAFNOSUPPORT)");
		return -EAFNOSUPPORT;
	}

	lock_sock(sk);

	switch (sock->state) {
	default:
		err = -EINVAL;
		goto out;
	case SS_CONNECTED:
		err = -EISCONN;
		goto out;
	case SS_CONNECTING:
		err = -EALREADY;
		/* already connecting - fall out of switch with EALREADY */
		break;
	case SS_UNCONNECTED:
		if (sk->sk_state != ESP_CLOSE) {
			err = -EISCONN;
			goto out;
		}

		/* do the actual connect */
		err = esp_connect(sk, uaddr, addr_len);
		if (err < 0)
			goto out;

		sock->state = SS_CONNECTING;

		/* non-blocking sockets will report EINPROGRESS */
		err = -EINPROGRESS;
		break;
	}

	/* wait until connection is established for blocking sockets */
	timeo = sock_sndtimeo(sk, flags & O_NONBLOCK);

	if (SK_STATE(sk, ESPF_SYN_SENT | ESPF_SYN_RECV)) {
		/* error code is set above */
		if (!timeo || !esp_connect_wait(sk, timeo))
			goto out;

		err = sock_intr_errno(timeo);
		if (signal_pending(current))
			goto out;
	}

	/* check if the connection attempt failed or timed out */
	if (sk->sk_state == ESP_CLOSE) {
		err = sock_error(sk) ? : -ECONNABORTED;
		sock->state = SS_UNCONNECTED;
		goto out;
	}

	/* connection established successfully */
	sock->state = SS_CONNECTED;
	err = 0;

	PR_DEBUG("connected esp socket %p to "NETH_FORMAT"-%d", sk,
	         NETH_HEX(sen->sen_addr), ntohs(sen->sen_port));
out:
	PR_DEBUG("<- sock_esp_connect");
	release_sock(sk);
	return err;
}


/*
 * POLL handler - handles select() and poll()
 *
 * Note that we don't need to lock the socket, as the upper poll layers
 * take care of normal races (between the test and the event) and we don't
 * go look at any of the socket buffers directly.
 */
static unsigned int sock_esp_poll(struct file *file, struct socket *sock,
                                  poll_table *wait)
{
	struct sock *sk = sock->sk;
	struct esp_opt *esp_sk = ESP_SK(sk);
	unsigned int mask;

	/*   PR_DEBUG("polling esp socket %p", sk); */

	poll_wait(file, sk->sk_sleep, wait);

	mask = 0;

	/* special handling for listening sockets */
	if (sk->sk_state == ESP_LISTEN) {
		if (!list_empty(&esp_sk->acc_queue))
			mask |= POLLIN | POLLRDNORM;
		goto out;
	}

	/* exceptional events? */
	if (sk->sk_err)
		mask |= POLLERR;

	/* shut down? */
	if (sk->sk_shutdown == SHUTDOWN_MASK || sk->sk_state == ESP_CLOSE)
		mask |= POLLHUP;

	if (sk->sk_shutdown & RCV_SHUTDOWN)
		mask |= POLLIN | POLLRDNORM;

	/* connected? */
	if (SK_STATE(sk, (ESPF_SYN_SENT | ESPF_SYN_RECV)))
		goto out;

	/* readable? */
	if (esp_sk->recv_seq != esp_sk->copy_seq || esp_sk->copy_off > 0)
		mask |= POLLIN | POLLRDNORM;

	/* writable? */
	if (!(sk->sk_shutdown & SEND_SHUTDOWN)) {
		if (sk_stream_wspace(sk) >= sk_stream_min_wspace(sk)) {
			mask |= POLLOUT | POLLWRNORM;
		} else {
			set_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);
			set_bit(SOCK_NOSPACE, &sk->sk_socket->flags);

			/* Race breaker. If space is freed after
			 * wspace test but before the flags are set,
			 * IO signal will be lost.
			 */
			if (sk_stream_wspace(sk) >= sk_stream_min_wspace(sk))
				mask |= POLLOUT | POLLWRNORM;
		}
	}

out:
	/*   PR_DEBUG("poll on esp socket %p returnes %#x", sk, mask); */
	return mask;
}


/*
 * GETNAME handler - handles getsockname() and getpeername()
 */
static int sock_esp_getname(struct socket *sock, struct sockaddr *uaddr,
                            int *uaddr_len, int peer)
{
	struct net_device *dev;
	struct sock *sk = sock->sk;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sockaddr_en *sen = (struct sockaddr_en *) uaddr;
	int err = 0;

	PR_DEBUG("-> sock_esp_getname");

	memset(sen, 0, sizeof(*sen));
	sen->sen_family  = AF_ENET;

	lock_sock(sk);

	if (peer) {
		if (sock->state != SS_CONNECTED) {
			err = -ENOTCONN;
			goto out_release;
		}
		sen->sen_ifindex = esp_sk->ifindex;
		sen->sen_port    = esp_sk->rport;
		sen->sen_addrlen = ETH_ALEN;
		memcpy(sen->sen_addr, esp_sk->raddr, ETH_ALEN);
	} else {
		sen->sen_ifindex = esp_sk->ifindex;
		sen->sen_port    = esp_sk->lport;
		dev = dev_get_by_index(INIT_NET_ARG esp_sk->ifindex);
		if (dev) {
			sen->sen_addrlen = ETH_ALEN;
			memcpy(sen->sen_addr, dev->dev_addr, ETH_ALEN);
			dev_put(dev);
		} else {
			sen->sen_addrlen = 0;
		}
	}
	*uaddr_len = sizeof(*sen);

out_release:
	release_sock(sk);

	PR_DEBUG("<- sock_esp_getname");

	return err;
}


/*
 * IOCTL handler
 */
static int sock_esp_ioctl(struct socket *sock, unsigned int cmd,
                          unsigned long arg)
{
	struct sock *sk = sock->sk;

	PR_DEBUG("ioctl command %#x  arg %#lx", cmd, arg);

	switch (cmd) {
	case SIOCOUTQ: {
		int amount = atomic_read(&sk->sk_wmem_alloc);
		return put_user(amount, (int __user *)arg);
	}
	case SIOCINQ: {
		struct sk_buff *skb;
		int amount = 0;

		spin_lock_bh(&sk->sk_receive_queue.lock);
		skb = skb_peek(&sk->sk_receive_queue);
		if (skb)
			amount = skb->len;
		spin_unlock_bh(&sk->sk_receive_queue.lock);
		return put_user(amount, (int __user *)arg);
	}
	case SIOCGSTAMP:
		return sock_get_timestamp(sk, (struct timeval __user *)arg);

	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}


/*
 * GETSOCKOPT handler
 */
static int sock_esp_getsockopt(struct socket *sock, int level, int optname,
                               char __user *optval, int __user *optlen)
{
	int len;
	struct sock *sk = sock->sk;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct net_device *dev;

	PR_DEBUG("getsockopt(%#x)", optname);

	if (level != SOL_EP)
		return -ENOPROTOOPT;

	if (get_user(len, optlen))
		return -EFAULT;

	if (len < 0)
		return -EINVAL;

	switch (optname) {
	case ESP_SO_MTU:
		/* check provided userspace buffer */
		if (len < sizeof(int))
			return -EINVAL;
		if (len > sizeof(int))
			len = sizeof(int);
		/* check if the socket is bound */
		if (!esp_sk->bound)
			return -EINVAL;
		/* calculate mtu */
		dev = dev_get_by_index(INIT_NET_ARG esp_sk->ifindex);
		if (dev) {
			unsigned int mtu = dev->mtu;
			mtu -= ESP_HLEN;

			PR_DEBUG("getsockopt determined %s mtu %u, esp mtu %d", dev->name, dev->mtu, mtu);

			dev_put(dev);
			if (put_user(mtu, (int __user *)optval))
				return -EFAULT;
		} else return -ENODEV;
		break;
	case ESP_SO_STATS: {
		struct esp_stats st;
		/* check provided userspace buffer */
		if (len < sizeof(st))
			return -EINVAL;
		len = sizeof(st);
		st = esp_sk->stats;
		memset(&esp_sk->stats, 0, sizeof(st));
		st.pkt_rx += st.pkt_rx_drop;

		if (copy_to_user(optval, &st, len))
			return -EFAULT;
		break;
	}
	default:
		return -ENOPROTOOPT;
	}

	/* announce size of the returned value to user */
	if (put_user(len, optlen))
		return -EFAULT;
	return 0;
}



/*
 * SHUTDOWN handler
 */
static int sock_esp_shutdown(struct socket *sock, int how)
{
	struct sock *sk = sock->sk;
	int err = 0;

	/* makes bit 1 rcvs and bit 2 snds */
	how++;

	if ((how & ~SHUTDOWN_MASK) || !how)     /* MAXINT->0 */
		return -EINVAL;

	lock_sock(sk);
	if (sock->state == SS_CONNECTING) {
		if (SK_STATE(sk, ESPF_SYN_SENT | ESPF_SYN_RECV | ESPF_CLOSE))
			sock->state = SS_DISCONNECTING;
		else
			sock->state = SS_CONNECTED;
	}

	switch (sk->sk_state) {
	case ESP_CLOSE:
		err = -ENOTCONN;
		break;
	case ESP_LISTEN:
		if (how & RCV_SHUTDOWN)
			esp_set_state(sk, ESP_CLOSE);
		break;
	default:
		sk->sk_shutdown |= how;
		if (how & SEND_SHUTDOWN)
			esp_sk_shutdown(sk);
		break;
	}

	/* wake up anyone sleeping on the socket */
	sk->sk_state_change(sk);
	release_sock(sk);
	return err;
}


/*
 * Release an ESP socket
 */
static int sock_esp_release(struct socket *sock)
{
	struct sock *sk = sock->sk;
	struct sk_buff *skb;
	long timeo = 0;
	int data_unread = 0;

	/* yes, this can really happen, e.g.
	 * when an accept() call on a listening socket was interrupted a
	 * child socket with no associated struct sock is the result... */
	if (!sk)
		return 0;

	PR_DEBUG("releasing esp socket %p in state %s", sk, esp_statenames[sk->sk_state]);

	lock_sock(sk);

	/* if linger is set this function should not return until the
	 * closing has completed
	 * if the close is due to the process exiting it should not linger
	 */
	if (sock_flag(sk, SOCK_LINGER) && !(current->flags & PF_EXITING))
		timeo = sk->sk_lingertime;

	/* purge receive queue */
	while ((skb = __skb_dequeue(&sk->sk_receive_queue)) != NULL) {
		data_unread += skb->len;
		__kfree_skb(skb);
	}

	if (data_unread) {
		/* we dropped some data unseen by the user -
		 * TODO: we should send a RST to inform the sender about that */
		PR_DEBUG("purged %d bytes of unread data from esp socket %p (rmem now: %d)",
		         data_unread, sk, atomic_read(&sk->sk_rmem_alloc));
		esp_set_state(sk, ESP_CLOSE);
	}

	/* shutdown the socket and send a FIN */
	esp_sk_shutdown(sk);

	PR_DEBUG("waiting for esp socket %p to close (timeout %lu)", sk, timeo);
	sk_stream_wait_close(sk, timeo);
	PR_DEBUG("State after wait_close for socket %p: %s", sk, esp_statenames[sk->sk_state]);

	/* last release_sock() in its lifetime - clears backlog receive queue */
	release_sock(sk);

	/* now the socket is in the kernel's hands */
	local_bh_disable();
	bh_lock_sock(sk);
	WARN_ON(sock_owned_by_user(sk));

	sock_hold(sk);
	/* now the socket is dead. No more input will appear */
	sock_orphan(sk); /* sets the SOCK_DEAD socket flag */
	sock->sk = NULL;

	if (sk->sk_state == ESP_CLOSE)
		esp_sk_destroy(sk);

	PR_DEBUG("esp socket %p released, refcount %d", sk,
	         atomic_read(&sk->sk_refcnt));

	/* release the socket */
	bh_unlock_sock(sk);
	local_bh_enable();
	sock_put(sk);
	return 0;
}

/*
 * ESP socket destructor
 */
void sock_esp_destruct(struct sock *sk)
{
	WARN_ON(atomic_read(&sk->sk_rmem_alloc));
	WARN_ON(atomic_read(&sk->sk_wmem_alloc));

	if (!sock_flag(sk, SOCK_DEAD)) {
		PR_ERR("attempt to release alive esp socket: %p", sk);
		return;
	}

	atomic_dec(&esp_skcount);

	PR_DEBUG("esp socket %p is free, %d are alive", sk, atomic_read(&esp_skcount));
}

/* ESP socket operations */
static struct proto_ops esp_ops = {
	.family     = PF_ENET,
	.owner      = THIS_MODULE,
	.release    = sock_esp_release,
	.bind       = sock_esp_bind,
	.connect    = sock_esp_connect,
	.socketpair = sock_no_socketpair,
	.accept     = sock_esp_accept,
	.getname    = sock_esp_getname,
	.poll       = sock_esp_poll,
	.ioctl      = sock_esp_ioctl,
	.listen     = sock_esp_listen,
	.shutdown   = sock_esp_shutdown,
	.setsockopt = sock_no_setsockopt,
	.getsockopt = sock_esp_getsockopt,
	.sendmsg    = sock_esp_sendmsg,
	.recvmsg    = sock_esp_recvmsg,
	.mmap       = sock_no_mmap,
	.sendpage   = sock_no_sendpage,
};

/*
 * Create an ESP socket
 */
struct sock *sock_esp_create(NET_PARAM struct socket *sock, int protocol, int *err)
{
	struct sock *sk;
	sk = esp_sk_alloc(NET_ARG GFP_KERNEL);
	if (!sk) {
		PR_ERR("Could not allocate esp kernel socket structure");
		*err = -ENOMEM;
		goto out;
	}

	/* init sock structure and esp private data */
	esp_sk_init(sock, sk);
	sk->sk_protocol = protocol;

	/* set operations to use with this socket (bind, send, recv, ...) */
	sock->ops = &esp_ops;

	atomic_inc(&esp_skcount);

	PR_DEBUG("ESP socket %p created, %d are alive.", sk, atomic_read(&esp_skcount));

out:
	return sk;
}

/*
 * Data receiving kernel hook function.
 * Here is where all the packets for our protocol arrive.
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,13)
/* In recent kernels we have a orig_dev parameter, which we ignore. */
static int esp_rcv_hook(struct sk_buff *skb, struct net_device *dev,
                        struct packet_type *pt, struct net_device *orig_dev)
#else
static int esp_rcv_hook(struct sk_buff *skb, struct net_device *dev,
                        struct packet_type *pt)
#endif
{
	struct sock *sk = NULL;
	struct ethhdr *eth_hdr;
	struct esp_header *esp_hdr;
	int data_len;
	int err = 0;
	ENET_DEBUG_DECLARE(struct timeval stamp);
	ENET_PROFILING_DECLARE(unsigned long long time[2]);

	/* take packet entry time */
	ENET_PROFILING_TIMESTAMP(&time[0]);

#if ENET_DEBUG
	skb_get_timestamp(skb, &stamp);
	PR_DEBUG("received esp frame on %s of length %d type %#x (ts %ld.%ld sh %d cl %d)",
	         dev->name, skb->len, ntohs(skb->protocol),
	         (long)stamp.tv_sec, (long)stamp.tv_usec,
	         skb_shared(skb), skb_cloned(skb));
#endif

	/* we dont want any packets of other protocols or for other hosts */
	if (skb->pkt_type != PACKET_HOST ||
	                skb->protocol != __constant_htons(ETH_P_ESP)) {
		PR_DEBUG("dropped esp frame for foreign protocol or destination host");
		goto out_free;
	}

	/* check if skb is shared and get a private header copy,
	 * because we will change and enqueue it */
	if ((skb = skb_share_check(skb, GFP_ATOMIC)) == NULL) {
		PR_DEBUG("could not allocate memory to clone a shared esp socket buffer");
		goto out_nomem;
	}
	eth_hdr = (struct ethhdr *) skb_mac_header(skb);

	/* check if there is enough data for at least the ESP header */
	if (!pskb_may_pull(skb, ESP_HLEN)) {
		PR_DEBUG("dropped too short esp frame");
		goto out_free;
	}

	/* pull the ESP header */
	__skb_pull(skb, ESP_HLEN);

	/* inspect ESP header, initialize SKB control data */
	esp_hdr = ESP_HDR(skb);
	ESP_SKB_CB(skb)->pkt_seq = ntohs(esp_hdr->pkt_seq);
	ESP_SKB_CB(skb)->ack_seq = ntohs(esp_hdr->ack_seq);
	data_len = ntohs(esp_hdr->length);

	PR_DEBUG("esp packet src "NETH_FORMAT"-%d dst "NETH_FORMAT"-%d len %d "
	         "seq %d ack %d "ESP_FLAGS_FORMAT,
	         NETH_HEX(eth_hdr->h_source), ntohs(esp_hdr->src_port),
	         NETH_HEX(eth_hdr->h_dest), ntohs(esp_hdr->dst_port),
	         data_len, ESP_SKB_CB(skb)->pkt_seq, ESP_SKB_CB(skb)->ack_seq,
	         ESP_FLAGS_DEBUG(esp_hdr->flags));

	/* trim the packet to the original length if the underlying medium
	 * padded it out */
	if (skb->len > data_len) {
		PR_DEBUG("trimming received esp packet to %u bytes", data_len);
		__pskb_trim(skb, data_len);
	}

	/* get the socket corresponding to / listening for a connection
	 * performs sock_hold(sk) */
	sk = esp_sk_lookup(eth_hdr->h_source, eth_hdr->h_dest,
	                   esp_hdr->src_port, esp_hdr->dst_port, dev->ifindex);

	/* take the skb off the device */
	skb->dev = NULL;
	/* but remember where it came from */
	SKB_INPUT_DEV(skb) = dev;

	if (!sk) {
		PR_DEBUG("no esp socket found for incoming packet");
		esp_send_rst(skb);
		goto out_free;
	}

	/* if the socket is in the user's hands we must not change its
	 * state - so we just append the received packet (that might be the
	 * cause of a state change) to its backlog queue, that will be
	 * processed via sk->sk_backlog_rcv() when the user releases the socket
	 */
	bh_lock_sock(sk);
	if (!sock_owned_by_user(sk)) {
		err = esp_rcv_packet(sk, skb);
	} else {
		PR_DEBUG("appending packet to esp socket %p backlog queue", sk);
		sk_add_backlog(sk, skb);
	}
	bh_unlock_sock(sk);

	sock_put(sk);

	/* take packet reception processing end time */
	ENET_PROFILING_TIMESTAMP(&time[1]);
	PR_PROFILING("RECV HOOK %llu - %llu (%llu)", time[0], time[1],
	             time[1] - time[0]);

	return err;

out_free:
	kfree_skb(skb);
out_nomem:
	return 0;
}


/*
 * Netdevice event notifier
 */
static int esp_dev_notifier(struct notifier_block *this, unsigned long msg,
                            void *data)
{
	struct sock *sk;
	struct net_device *dev = (struct net_device*)data;
	struct hlist_node *node;
	int bucket;

	read_lock(&esp_bhash_lock);
	HTABLE_SK_FOR_EACH_BOUND(sk, node, bucket, esp_bhash, ESP_BHTABLE_SIZE) {
		struct esp_opt *esp_sk = ESP_SK(sk);

		switch (msg) {
			/* notify sockets of dead devices */
		case NETDEV_UNREGISTER:
		case NETDEV_DOWN:
			if (dev->ifindex == esp_sk->ifindex) {
				if (sk_bound(sk)) {
					// close the socket
					SK_ERR_REPORT(sk, -ENETDOWN);
					esp_set_state(sk, ESP_CLOSE);
				}
			}
			break;
		}
	}

	read_unlock(&esp_bhash_lock);

	return NOTIFY_DONE;
}

/*
 * Kernel hook data structures
 */
struct packet_type esp_receive_hook = {
	.type           = __constant_htons(ETH_P_ESP), /**< enter the packet type hashtable */
	.dev            = NULL,              /**< get packets from all devices */
	.func           = esp_rcv_hook,      /**< packet receiving handler function */
	.af_packet_priv = NULL,              /**< private data */
};


struct notifier_block esp_netdev_notifier = {
	.notifier_call = esp_dev_notifier,   /**< netdevice event notifier */
};


/* init the esp protocol */
int esp_prot_init(void)
{
	int rc = 0;
	struct sk_buff *dummy_skb;

	/* esp skb control data size check */
	if (sizeof(struct esp_skb_cb) > sizeof(dummy_skb->cb)) {
		PR_ERR("esp skb control data structure too big (%d/%d bytes)",
		       (int)sizeof(struct esp_skb_cb),
		       (int)sizeof(dummy_skb->cb));
		rc = -EFAULT;
		goto out;
	}

	rc = esp_prot_init_slab();

	/* initialize the ACK queue */
	skb_queue_head_init(&esp_ack_queue);

out:
	return rc;
}


/* destroy the esp protocol */
void esp_prot_destroy(void)
{
	esp_prot_destroy_slab();
}


/*
 * Initialize some values of a new esp socket
 */
void esp_sk_init(struct socket *sock, struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* init the kernel socket structure */
	sock_init_data(sock, sk);

	sk->sk_family = PF_ENET;
	ESP_SK_USES_WRITE_QUEUE(sk);
	/* set kernel sock destructor */
	sk->sk_destruct = sock_esp_destruct;
	/* set backlog processing function */
	sk->sk_backlog_rcv = esp_rcv_packet;
	/* set writespace indication function */
	sk->sk_write_space = sk_stream_write_space;
	/* initial rcvbuf size */
	sk->sk_rcvbuf = esp_default_rcvbuf;
	/* initial sndbuf size */
	sk->sk_sndbuf = esp_default_sndbuf;

	/* init lists */
	INIT_LIST_HEAD(&esp_sk->syn_queue);
	INIT_LIST_HEAD(&esp_sk->acc_queue);
	skb_queue_head_init(&esp_sk->oos_packets);

	/* init RTX and RRQ timers */
	init_timer(&esp_sk->rtx_timer);
	esp_sk->rtx_timer.function = &esp_rtx_timer;
	esp_sk->rtx_timer.data = (unsigned long) sk;

	init_timer(&esp_sk->rrq_timer);
	esp_sk->rrq_timer.function = &esp_rrq_timer;
	esp_sk->rrq_timer.data = (unsigned long) sk;
}
