/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#ifndef ENET_STREAM_H_
#define ENET_STREAM_H_

#include <linux/version.h> /* kernel version dependency */
#include <net/sock.h>  /* struct sock, basic socket operations */
#include <linux/skbuff.h> /* socket buffer struct and ops */
#include <linux/timer.h> /* struct timer_list and timer ops */
#include <linux/tcp.h>  /* tcp socket state constants */

#include "enet_module.h"
#include "enet_compat.h"

/*
 * Ethernet packet type to use for ESP frames
 *
 * (choose a packet type currently unused by the kernel, see
 *  ETH_P_* constants in %KERNEL_SRC%/include/linux/if_ether.h
 *  or http://www.iana.org/assignments/ethernet-numbers)
 *
 * Furthermore, the packet type should be larger than or equal to
 * 0x0600 (1536). Otherwise the linux kernel assumes that this
 * value denotes the frame length of an IEEE-802.3 ethernet frame,
 * which have a slightly different structure compared to the
 * 'normal' blue-book ethernet frame.
 */
#define ETH_P_ESP 0x8887

/* ESP packet header */
struct esp_header {
	unsigned short dst_port; /**< esp destination port */
	unsigned short src_port; /**< esp source port */
	unsigned short pkt_seq;  /**< esp packet sequence number */
	unsigned short ack_seq;  /**< acknowledgement sequence number */
	unsigned short length;   /**< data length */
	unsigned char flags;     /**< esp flags */
} __attribute__((packed));


/* ESP header length */
#define ESP_HLEN sizeof(struct esp_header)

/* ESP packet flags */
enum __esp_flags {
	ESP_FLAG_SYN = 1 << 0, /**< synchronize on connection startup */
	ESP_FLAG_ACK = 1 << 1, /**< acknowledge received packets */
	ESP_FLAG_FIN = 1 << 2, /**< announce connection closure */
	ESP_FLAG_RST = 1 << 3, /**< request connection reset */
	ESP_FLAG_RRQ = 1 << 4, /**< retransmission request */
	ESP_FLAG_TXS = 1 << 5, /**< announce start of transmission */
	ESP_FLAG_TXF = 1 << 6, /**< announce transmission is finished */
	ESP_FLAG_XXX = 1 << 7  /**< unused */
};

#define ESP_FLAGS_FORMAT "[ %s%s%s%s%s%s%s]"

#define ESP_FLAGS_DEBUG(flags)   \
 (flags & ESP_FLAG_SYN)? "SYN " : "", \
 (flags & ESP_FLAG_ACK)? "ACK " : "", \
 (flags & ESP_FLAG_FIN)? "FIN " : "", \
 (flags & ESP_FLAG_RST)? "RST " : "", \
 (flags & ESP_FLAG_RRQ)? "RRQ " : "", \
 (flags & ESP_FLAG_TXS)? "TXS " : "", \
 (flags & ESP_FLAG_TXF)? "TXF " : ""

/* Private ESP socket options */
struct esp_opt {
	struct esp_stats stats;        /**< esp socket statistics */

	/* bind management */
	unsigned char bound;             /**< whether the socket is bound */
	int ifindex;                     /**< bound interface */
	unsigned short port;             /**< bound port (host byte order) */
	unsigned short lport;     /**< bound source port (net byte order) */
	unsigned char laddr[ETH_ALEN];   /**< local interface address */

	/* connection management */
	unsigned short rport;     /**< connected destination port (net byte order) */
	unsigned char raddr[ETH_ALEN];   /**< connected destination address */
	struct sock *parent;             /**< parent socket during connection establishing */
	unsigned short synq_len;     /**< syn queue length */
	struct list_head syn_queue;      /**< queue of half-connected children (SYN_SENT/RECV) */
	struct list_head acc_queue;      /**< queue of connected children waiting for accept() */
	unsigned short rcv_mem_pressure; /**< this socket has low rcv. memory */
	unsigned short trans_announced;  /**< if the TXS flag was already sent out */
	unsigned short announce_seq;     /**< # of the last pkt. sent with TXS flag set */
	unsigned short recv_announced;   /**< if we received a TXS flagged packet */
	unsigned short need_rtx;         /**< if we need a retransmission once memory gets freed */

	unsigned char rto_scale;

	/* sequencing management */
	unsigned short send_seq;        /**< sequence number next to send */
	unsigned short recv_seq;        /**< sequence number expected next */
	unsigned short una_seq;        /**< seq. # of the first sent but un-ACK'ed packet */
	unsigned short ack_seq;        /**< least sequence number to ACK next */
	struct timer_list rrq_timer;     /**< retransmission request timer */
	struct sk_buff_head oos_packets;    /**< out-of-sequence packet queue */
	unsigned int skb_size;        /**< size of received socket buffers for memory calculations */

	/* retransmission management */
	unsigned long srtt;        /**< smoothed round-trip-time estimation */
	unsigned long rto;        /**< dynamic retransmission timeout */
	struct timer_list rtx_timer;     /**< retransmission timer */

	/* user data copy management */
	unsigned short copy_seq;        /**< seqence number of packet to copy next to user */
	unsigned short copy_off;        /**< offset inside the packet next to copy */
};


/* ESP socket data structure */
struct esp_sock {
	struct sock sk;      /**< basic kernel socket struct */
	struct esp_opt esp;  /**< esp socket options */
};


/* access private ESP socket information */
#define ESP_SK(__sk) \
 (&((struct esp_sock *)(__sk))->esp)

#define SK(__esp_sk) \
 ((struct sock *)container_of(__esp_sk, struct esp_sock, esp))


/* counter of ESP sockets alive */
extern atomic_t esp_skcount;

/* ESP socket states
 * (TCP socket states are used in some basic socket functions, so
 *  our constants have to be equivalent to them)
 */
enum __esp_sk_states {
	ESP_ESTABLISHED = TCP_ESTABLISHED,
	ESP_SYN_SENT    = TCP_SYN_SENT,
	ESP_SYN_RECV    = TCP_SYN_RECV,
	ESP_FIN_WAIT1   = TCP_FIN_WAIT1,
	ESP_FIN_WAIT2   = TCP_FIN_WAIT2,
	ESP_TIME_WAIT   = TCP_TIME_WAIT,
	ESP_CLOSE       = TCP_CLOSE,
	ESP_CLOSE_WAIT  = TCP_CLOSE_WAIT,
	ESP_LAST_ACK    = TCP_LAST_ACK,
	ESP_LISTEN      = TCP_LISTEN,
	ESP_CLOSING     = TCP_CLOSING,
	ESP_MAX_STATES
};

enum __esp_sk_state_flags {
	ESPF_ESTABLISHED = (1 << ESP_ESTABLISHED),
	ESPF_SYN_SENT    = (1 << ESP_SYN_SENT),
	ESPF_SYN_RECV    = (1 << ESP_SYN_RECV),
	ESPF_FIN_WAIT1   = (1 << ESP_FIN_WAIT1),
	ESPF_FIN_WAIT2   = (1 << ESP_FIN_WAIT2),
	ESPF_TIME_WAIT   = (1 << ESP_TIME_WAIT),
	ESPF_CLOSE       = (1 << ESP_CLOSE),
	ESPF_CLOSE_WAIT  = (1 << ESP_CLOSE_WAIT),
	ESPF_LAST_ACK    = (1 << ESP_LAST_ACK),
	ESPF_LISTEN      = (1 << ESP_LISTEN),
	ESPF_CLOSING     = (1 << ESP_CLOSING),
};


/* check for the socket being in one of the OR'ed states */
#define SK_STATE(_sk, arg) ((1 << _sk->sk_state) & (arg))


/* syn queue management functions */
static inline void esp_synq_added(struct sock *sk)
{
	++ESP_SK(sk)->synq_len;
}


static inline void esp_synq_removed(struct sock *sk)
{
	--ESP_SK(sk)->synq_len;
}


static inline int esp_synq_is_full(struct sock *sk)
{
	return ESP_SK(sk)->synq_len +
	       sk->sk_ack_backlog > sk->sk_max_ack_backlog;
}

struct esp_header;

/* ESP socket buffer control block */
struct esp_skb_cb {
	unsigned short pkt_seq; /**< packet sequence number */
	unsigned short ack_seq; /**< ack sequence number */
	unsigned long  sent_ts; /**< sent timestamp for rtt measuring */
	unsigned long  timeo;   /**< retransmission timeout */
	unsigned short xmits;   /**< number of (re)transmissions */
};

/* access the private socket buffer control block data */
#define ESP_SKB_CB(__skb) ((struct esp_skb_cb *)&((__skb)->cb[0]))

/*
 * ESP ports
 *
 * There's several hash tables containing ESP sockets:
 * - one for bound sockets maintaining the socket-to-port relationship
 *   (here the socket lookup is performed just by the local port;
 *    it is also used to determine allocated ports)
 * - one for listening sockets
 *   (here the socket lookup is performed just by the local port;
 *    it is used to speed-up the lookup for connection requests)
 * - one for sockets with established connection or etablishing a connection
 *   (here the socket lookup is performed by the local and remote
 *    ports and addresses)
 * Any bound socket can be found in the bound sockets hashtable.
 * Listening or connected sockets are kept additionally in the
 * corresponding hashtable.
 */

/* port-to-socket hashtable size
 * (define a hashtable size of 2^x - it is often used like:  x & (ESP_?HTABLE_SIZE - 1) ) */
#define ESP_BHTABLE_SIZE 128
#define ESP_LHTABLE_SIZE 128
#define ESP_EHTABLE_SIZE 128

/* port-to-socket hashtable */
extern struct hlist_head esp_bhash[ESP_BHTABLE_SIZE];
extern struct hlist_head esp_lhash[ESP_LHTABLE_SIZE];
extern struct hlist_head esp_ehash[ESP_EHTABLE_SIZE];

/* port-to-socket hashtable r/w-lock */
extern rwlock_t esp_bhash_lock;
extern rwlock_t esp_lhash_lock;
extern rwlock_t esp_ehash_lock;

/* port range to automatically choose ports from and port rover */
extern int esp_local_port_range[2];
extern int esp_port_rover;

/* hash function for bound sockets (works on host byte order port numbers) */
#define ESP_BHASH_FN(lport) ((lport) & (ESP_BHTABLE_SIZE - 1))

/* determines whether a given port is already in use */
static inline int esp_port_inuse(u16 lport)
{
	struct sock *sk;
	struct hlist_node *node;

	sk_for_each_bound(sk, node, &esp_bhash[ESP_BHASH_FN(lport)]) {
		if (ESP_SK(sk)->port == lport)
			return 1;
	}

	return 0;
}

/* hash function for listening sockets
 * folds the port to achieve more noise independent from the byte order */
static inline u16 esp_lhash_fn(u16 lport)
{
	return ((lport ^(lport >> 8)) & (ESP_LHTABLE_SIZE - 1));
}

/* hash functions for connected sockets */
static inline u16 esp_ehash_fn(char *laddr, char *raddr, u16 lport, u16 rport)
{
	u16 hash =
	        *((u16*) & laddr[0]) ^ *((u16*) & laddr[2]) ^ *((u16*) & laddr[4]) ^
	        *((u16*) & raddr[0]) ^ *((u16*) & raddr[2]) ^ *((u16*) & raddr[4]) ^
	        lport ^ rport;
	hash ^= hash >> 8;

	return (hash & (ESP_EHTABLE_SIZE - 1));
}

/*
 * ESP Data retransmission
 *
 * Packet retransmission can be controlled in two ways:
 * (1) A maximum amount of retransmissions can be specified after which the
 *     connection is considered dead.
 * (2) The retransmission timeout can be doubled after each retransmission
 *     attempt. If it exceeds the maximum timeout value the connection
 *     is considered dead again.
 *     This results in a (very rudimentary) flow control.
 * At least one option must be enabled to avoid infinite waiting, but it is
 * possible to use both concurrently.
 */

/* whether the timeout should be doubled after each retransmission attempt */
#define ESP_DOUBLE_RTO 0


/* whether the connection should be considered dead if the retransmission
 * timeout of the socket reached its maximum (0/1) - requires ESP_DOUBLE_RTO */
#define ESP_TIMEOUT_CONN 0


/* the amount of retransmission attempts of a single packet after which the
 * connection should be considered dead (zero means unlimited) */
#define ESP_MAX_RETRANSMITS 0


/* whether to use an additional queue for packets received out-of-sequence */
#define ESP_USE_OOS_QUEUE 0


/* the measure of the actual time */
#define ESP_TIMESTAMP ((unsigned long)jiffies)

/* retransmission timeout calculation constants */
#define ESP_RESOURCE_TO ((unsigned long)(HZ/20))
#define ESP_RTO_INIT    ((unsigned long)(HZ/10))
#define ESP_MAX_RTO     ((unsigned long)(HZ))

#if HZ < 1000
#define ESP_MIN_RTO     ((unsigned long)(2))
#else
#define ESP_MIN_RTO     ((unsigned long)(HZ/500))
#endif


/* maximum linger time in the FINWAIT-2 state */
#define ESP_FINWAIT2_TO ((unsigned long)(60*HZ))

/* the maximum amount of packets to receive before an ACK is sent */
#define ESP_MAX_ACK_PENDING 10

/* the maximum time to wait before an ACK is sent (if one is pending) */
#define ESP_ACK_DELAY (2 * ESP_MIN_RTO / 3)

/* calculate the smoothed round-trip-time and retransmission timeout */
static inline void esp_update_rto(struct sock *sk, unsigned long val)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	if (esp_sk->srtt) {
		/* rtt has already been measured, so take
		 * srtt = 7/8 srtt + 1/8 val */
		esp_sk->srtt += ((long)val - (long)esp_sk->srtt) >> 3;
	} else {
		/* rtt has not yet been measured, so take val as rtt */
		esp_sk->srtt = val;
	}

	/* update the rtx timeout including some tolerance T, so that
	 * rto = T * srtt   and
	 * ESP_MIN_RTO <= rto <= ESP_MAX_RTO */
	esp_sk->rto = esp_sk->srtt << 1;
	esp_sk->rto = min(ESP_MAX_RTO, max(ESP_MIN_RTO, esp_sk->rto));

	PR_DEBUG("updated esp socket %p srtt %lu rto %lu by val %lu (jiffies %lu)",
	         sk, esp_sk->srtt, esp_sk->rto, val, ESP_TIMESTAMP);
}

void esp_rtx_timer_reset(struct sock *sk, unsigned long when);
void esp_sk_init(struct socket *sock, struct sock *sk);
struct sock *esp_sk_spawn_child(struct sock *parent, struct sk_buff *skb);

/* double the retransmission timeout (if configured) */
static inline void esp_double_rto(struct sock *sk)
{
	if (ESP_DOUBLE_RTO) {
		struct esp_opt *esp_sk = ESP_SK(sk);

		esp_sk->rto <<= ESP_DOUBLE_RTO;
		if (esp_sk->rto > ESP_MAX_RTO)
			esp_sk->rto = ESP_MAX_RTO;

		PR_DEBUG("increased esp socket %p srtt %lu rto %lu (jiffies %lu)",
		         sk, esp_sk->srtt, esp_sk->rto, ESP_TIMESTAMP);
	}
}

/*
 * convenience function for skb allocation
 */
static inline struct sk_buff *esp_skb_alloc(struct net_device *dev, void *daddr,
			                size_t data_len, gfp_t allocation, int *err) {
	struct sk_buff *skb;

	skb = alloc_skb(LL_RESERVED_SPACE(dev) + ESP_HLEN + data_len, allocation);
	if (!skb) {
		*err = -ENOBUFS;
		goto out;
	}

	skb_reserve(skb, LL_RESERVED_SPACE(dev));
	skb_reset_network_header(skb);

	/* copy ethernet header into skb (function performs htons(type) ) */
	if (HAS_HARD_HEADER(dev)) {
		if (dev_hard_header(skb, dev, ETH_P_ESP, daddr, NULL, 0) < 0) {
			*err = -EINVAL;
			goto out_free;
		}
	}

	/* set skb properties */
	skb->dev      = dev;
	skb->protocol = __constant_htons(ETH_P_ESP);

	/* init socket buffer control data */
	ESP_SKB_CB(skb)->xmits = 0;

out:
	return skb;

out_free:
	kfree_skb(skb);
	return NULL;
}

/*
 * Stops the timer for RRQs
 */
static inline void esp_rrq_clear(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* stop the timer */
	sk_stop_timer(sk, &esp_sk->rrq_timer);

	PR_DEBUG("Stopped RRQ timer on socket %p.", sk);
}


/* compare two 16 bit unsigned ints and consider wraparound */
static inline int seq_before(u16 seq1, u16 seq2)
{
	return (s16)(seq1 -seq2) < 0;
}


static inline int seq_after(u16 seq1, u16 seq2)
{
	return (s16)(seq2 -seq1) < 0;
}

/* determine whether ACKs are pending on a socket */
#define ESP_ACK_PENDING(__sk) \
 (seq_after(ESP_SK(__sk)->recv_seq,ESP_SK(__sk)->ack_seq))


/* reset delayed ACK timer */
#if ENET_DEBUG && 0 /* "&& 0" to stop these messages! */
#define ESP_RESET_RRQ_TIMER(__sk, __when) \
        do { PR_DEBUG("Resetting RRQ timeout to %lu.", __when); \
             sk_reset_timer(__sk, &ESP_SK(__sk)->rrq_timer, ESP_TIMESTAMP + (__when)); \
           } while (0)
#else
#define ESP_RESET_RRQ_TIMER(__sk, __when) \
 sk_reset_timer(__sk, &ESP_SK(__sk)->rrq_timer, ESP_TIMESTAMP + (__when))
#endif

/*
 * ESP socket operation prototypes
 */
int esp_prot_init(void);
void esp_prot_destroy(void);
struct sock *sock_esp_create(NET_PARAM struct socket *sock, int protocol, int *err);
void esp_openreq_enqueue(struct sock *parent, struct sock *child);
void esp_openreq_dequeue(struct sock *child);

/*
 * ESP kernel hook data structures
 */
extern struct packet_type    esp_receive_hook;
extern struct notifier_block esp_netdev_notifier;

#ifdef CONFIG_PROC_FS
extern struct file_operations esp_proc_fops;
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11)

extern struct proto esp_proto;

/* register the esp protocol and create the socket slab cache */
static inline int esp_prot_init_slab(void)
{
	return proto_register(&esp_proto, 1);
}

/* unregister protocol esp and destroy socket slab cache */
static inline void esp_prot_destroy_slab(void)
{
	proto_unregister(&esp_proto);
}

/* esp socket allocation function */
static inline struct sock *esp_sk_alloc(NET_PARAM gfp_t priority) {
	return sk_alloc(NET_ARG PF_ENET, priority, &esp_proto);
}

#define ESP_SK_USES_WRITE_QUEUE(__sk) \
 sock_set_flag(__sk, SOCK_USE_WRITE_QUEUE)



#else

#endif

#endif /*ENET_STREAM_H_*/
