/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#ifndef ENET_STREAM_CONG_H_
#define ENET_STREAM_CONG_H_

#include "enet_stream.h"
#include "enet_stream_output.h"
#include "enet_compat.h"


/**
 * This timeout is used for the RTX timer,
 * which is needed only if the TXS flagged
 * packet is lost. Very unlikely.
 */
#define ESP_LONG_TIMEOUT (HZ * 20)

/**
 * Maximum # of packets to xmit at once (send window size).
 */
extern int esp_burst_length;

/**
 * Initial burst length
 */
#define ESP_INITIAL_BURST_LENGTH 21

/**
 * Burst length while waiting for the initial ACK.
 */
extern int esp_initial_ack_burst_length;
#define ESP_INITIAL_ACK_INITIAL_BURST_LENGTH 12

/**
 * Number of packets to ack at once.
 */
extern int esp_packets_to_ack;

/**
 * Initial value for packets to ack.
 */
#define ESP_INITIAL_PACKETS_TO_ACK 10

/**
 * The round trip time
 */
extern int esp_round_trip_time;

/**
 * The initial round trip time used for all connections
 */
#define ESP_INITIAL_ROUND_TRIP_TIME (HZ / 20)

/**
 * Initial size of the receive buffer
 */
#define ESP_INITIAL_RCVBUF (1024 * 1024)

/**
 * Initial size of the send buffer
 */
#define ESP_INITIAL_SNDBUF (1024 * 1024)

/**
 * Number of connections we currently receive data on.
 */
extern atomic_t esp_recv_count;

/**
 * This is the queue where ACK packets are parked if
 * we are receiving from more than one host or to avoid
 * congestion.
 */
extern struct sk_buff_head esp_ack_queue;

/**
 * For making the default sk_rcvbuf value accessible via
 * sysctl and independent from the value used for TCP sockets.
 */
extern int esp_default_rcvbuf;

/**
 * For making the default sk_sndbuf value accessible via
 * sysctl and independent from the value used for TCP sockets.
 */
extern int esp_default_sndbuf;

/**
 * For sysctl support.
 */
extern struct ctl_table esp_ctl_table[];


/**
 * For walking the ack queue. Be sure to grab esp_ack_queue.lock
 * if necessary.
 */
#define __esp_walk_ack_queue(__skb) for (__skb = esp_ack_queue.next;   \
                      prefetch(skb->next), __skb != (struct sk_buff *)&esp_ack_queue; \
                      __skb = __skb->next)

/**
 * Tells if we are currently waiting for the first ACK (of
 * a new transmission).
 *
 * @return 0 if false, other value otherwise
 */
static inline unsigned int esp_waiting_for_first_ack(struct sock *sk)
{
	const struct esp_opt *esp_sk = ESP_SK(sk);
	return !seq_after(esp_sk->una_seq, esp_sk->announce_seq);
}

/**
 * Tells if it's ok to xmit the given skb.
 * @return 1 if it's ok to do so, 0 otherwise
 */
static inline unsigned int esp_cwnd_test(struct sock *sk,
                struct sk_buff *skb)
{
	const struct esp_opt *esp_sk = ESP_SK(sk);
	int res = seq_before(ESP_SKB_CB(skb)->pkt_seq,
	                     esp_sk->una_seq + esp_burst_length);

	/* Check if the first data packet (carrying the TXS flag) was
	 * surely received. (If not we may emit only
	 * esp_initial_ack_burst_length packets)
	 */
	if (!esp_waiting_for_first_ack(sk)) {
		/* it was, we may have up to esp_burst_length packets in flight */
		PR_DEBUG("cwnd_test = %d (pkt_seq %d, una_seq %d, goal %d)",
		         res, ESP_SKB_CB(skb)->pkt_seq, esp_sk->una_seq,
		         esp_burst_length);
	} else {
		/* it wasn't, only packets up to (including) the one carrying
		 * the TXS flag may be sent.
		 */
		res &= !seq_after(ESP_SKB_CB(skb)->pkt_seq, esp_sk->announce_seq +
		                  esp_initial_ack_burst_length);
		PR_DEBUG("cwnd_test = %d (waiting for TXS ACK, pkt_seq %d, una_seq %d, txs_seq %d)", res,
		         ESP_SKB_CB(skb)->pkt_seq, esp_sk->una_seq, esp_sk->announce_seq);
	}

	return res;
}

/**
 * Checks if there's enough recv. mem available to store
 * a full burst of packets.
 */
static inline int esp_socket_low_rcv_mem(struct sock *sk)
{
	const struct esp_opt *esp_sk = ESP_SK(sk);
	return ((atomic_read(&sk->sk_rmem_alloc) +
	         2 * esp_burst_length * esp_sk->skb_size) >= sk->sk_rcvbuf);
}

/**
 * Checks if there's enough recv. mem available to lift
 * the mem. pressure situation. (Checks if at least half
 * of rcv. mem is free now.)
 */
static inline int esp_socket_mem_press_lifted(struct sock *sk)
{
	return !esp_socket_low_rcv_mem(sk) &&
	       (atomic_read(&sk->sk_rmem_alloc) <= (sk->sk_rcvbuf >> 1));
}

/**
 * Determines if it's time to send out an ACK for the data just
 * received.
 *
 * @return 1 if an ACK should be sent, 0 otherwise
 */
static inline int esp_should_send_ack(const struct esp_opt *esp_sk)
{
	const int res = !seq_after(esp_sk->ack_seq + esp_packets_to_ack,
	                           esp_sk->recv_seq);
	PR_DEBUG("should send ack = %d (ack_seq %d, recv_seq %d, goal %d)",
	         res, esp_sk->ack_seq, esp_sk->recv_seq, esp_packets_to_ack);
	return res;
}

/**
 * Finds the ACK owned by the given socket on the ACK queue,
 * dequeues and returns it. If no ACK for the socket is
 * found on the queue NULL is returned.
 *
 * @return the skb containing the ACK for the socket
 *         or NULL if it wasn't there
 */
static inline struct sk_buff *esp_dequeue_ack(struct sock *sk) {
	struct sk_buff *skb = NULL;

	spin_lock_bh(&esp_ack_queue.lock);

	__esp_walk_ack_queue(skb) {
		const struct sock *sk1 = skb->sk;
		/* all the ACKs on queue MUST be owned by a socket */
		WARN_ON(!sk1);
		if (sk1 == sk) {
			/* we found the ACK in question */
			__skb_unlink(skb, &esp_ack_queue);
			goto out_found;
		}
	}

	skb = NULL;

out_found:
	spin_unlock_bh(&esp_ack_queue.lock);
	return skb;
}

/**
 * Purges the ACK of the given socket from the ACK queue.
 * @return 1 if an ACK was found and purged,
 *         0 otherwise
 */
static inline int esp_purge_ack(struct sock *sk)
{
	struct sk_buff *skb = esp_dequeue_ack(sk);

	PR_DEBUG("purged ACK for socket %p\n", sk);

	if (skb) {
		__kfree_skb(skb);
		return 1;
	}

	return 0;
}

/**
 * Shrinks the ACK queue to a reasonable size.
 *
 * @return the # of ACKs found and xmitted,
 *         0  if no ACK was there
 */
static inline int esp_shrink_ack_queue(void)
{
	struct sk_buff *skb = NULL;
	int res = 0;
	spin_lock_bh(&esp_ack_queue.lock);

	while (!(esp_ack_queue.qlen < atomic_read(&esp_recv_count)) &&
	                atomic_read(&esp_recv_count)) {
		skb = __skb_dequeue(&esp_ack_queue);
		PR_DEBUG("shrink ack queue, sending ack for: %p\n", skb->sk);
		res += 1;
		ESP_RESET_RRQ_TIMER(skb->sk, esp_round_trip_time);
		dev_queue_xmit(skb);
		/* FIXME: No error handling here */
	}

	spin_unlock_bh(&esp_ack_queue.lock);
	return res;
}

void esp_enter_rcv_mem_pressure(struct sock *sk);
void esp_lift_rcv_mem_pressure(struct sock *sk);

#endif
