/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#ifndef ENET_STREAM_INPUT_H_
#define ENET_STREAM_INPUT_H_

#include "enet_module.h"
#include "enet_stream_common.h"
#include "enet_stream_output.h"
#include "enet_stream_cong.h"
#include "enet_compat.h"

/**
 * Handles data packets with TXS / TXF flags set. Must be called after the
 * packet in question has been enqueued onto receive buffer to avoid sending
 * bogus ACKs, as I learned today.
 *
 * @return 1 if an ACK needs to be enqueued, 0 if everything's fine.
 */
static inline int esp_process_txs_txf(struct sock *sk, struct sk_buff *skb)
{
	struct esp_header *esp_hdr = ESP_HDR(skb);
	struct esp_opt *esp_sk = ESP_SK(sk);

	if ((esp_hdr->flags & ESP_FLAG_TXS) &&
	                (esp_hdr->flags & ESP_FLAG_TXF)) {
		/* all the data was sent in one frame, we can skip
		 * most of the handling */

		esp_send_ack(sk);

		return 0;
	}

	/* check for transmission announcement */
	if (esp_hdr->flags & ESP_FLAG_TXS) {
		if (!esp_sk->recv_announced) {
			esp_sk->recv_announced = 1;
			/* oh, it's new for us that this peer want's
			 * to send us some data, check if we can use
			 * it now.
			 */
			PR_DEBUG("Socket %p processes a TXS\n", sk);
			atomic_inc(&esp_recv_count);

			if (!esp_socket_low_rcv_mem(sk)) {
				PR_DEBUG("Transmission started on socket %p, now: %d.",
				         sk, atomic_read(&esp_recv_count));
				return 1;
			} else {
				/* Sadly, we can't take more data at this time.
				 * Enter memory pressure mode. This will be
				 * lifted as the user reads some data from our
				 * read queue.
				 */
				esp_enter_rcv_mem_pressure(sk);
				return 0;
			}
		} else {
			PR_DEBUG("Bogus TXS flag. Retransmission?");
			/* Anyway, we should ACK this packet soon to make
			 * the sender calm down. */
			return 1;
		}
	}

	/* check for transmission end */
	if (esp_hdr->flags & ESP_FLAG_TXF) {
		if (esp_sk->recv_announced) {
			if (!esp_sk->rcv_mem_pressure)
				atomic_dec(&esp_recv_count);

			PR_DEBUG("Transmission finished on socket %p, now: %d.",
			         sk, atomic_read(&esp_recv_count));

			esp_sk->recv_announced = 0;
			esp_sk->rcv_mem_pressure = 0;
			esp_sk->need_rtx = 0;

			/* Send out an ACK. We should NOT send the ACK we
			 * probably have on queue, as it's outdated for
			 * sure (we just received new data, didn't we?).
			 * It's OK to send an ACK immediately as TXF means
			 * the sender is finished anyway.
			 *
			 * FIXME: This sends a duplicate ACK if pkts_to_ack == 1, as
			 * the esp_rcv_packet function will have sent this ACK
			 * already.
			 */
			esp_purge_ack(sk);
			esp_send_ack(sk);
			esp_shrink_ack_queue();

			/* Additionally we may clear the RRQ timer */
			esp_rrq_clear(sk);
		} else {
			PR_DEBUG("Bogus TXF flag. Retransmission?");
			/* We really wish the sender would calm down... */
			esp_send_ack(sk);
		}
	}

	return 0;
}

/**
 * Processes a received retransmission request.
 * MUST be called ONLY after process_ack was called because it relies on
 * the last packet on the write queue is the first to be retransmitted.
 * The skb_orig parameter is currently unused.
 */
static inline int esp_process_rrq(struct sock *sk, struct sk_buff *skb_orig)
{
	//struct esp_opt *esp_sk = ESP_SK(sk);
	struct sk_buff *skb = sk->sk_write_queue.next;
	int cnt = 0;

	printk("socket %p processes an RRQ (wq len: %d)\n",
	       sk, sk->sk_write_queue.qlen);

	if (sk->sk_write_queue.next == (struct sk_buff *)(&sk->sk_write_queue)) {
		printk("Bogus RRQ received, write queue is empty.\n");
		return 0;
	}

	do {
		/* we may push frames up to (including) the send head */
		if ((skb == (struct sk_buff *)&sk->sk_write_queue))
			break;

		esp_skb_xmit(sk, skb_clone(skb, GFP_ATOMIC));
		++cnt;

		if (skb == sk->sk_send_head) break;
		skb = skb->next;
	} while (1);

	printk("socket %p processed RRQ, sent out %d frames again.\n", sk, cnt);

	/* and now we push what else we are allowed to ... */
	esp_push_pending_frames(sk);

	return 1;
}

/**
 * Checks if the ACK is usefull and if it is, adjusts the una_seq.
 * @return 1 if it did so, 0 otherwise
 */
static inline int esp_process_ack_early(struct sock *sk, struct sk_buff *skb)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* check if there is a packet waiting for this ACK */
	if (!seq_after(ESP_SKB_CB(skb)->ack_seq, esp_sk->una_seq) ||
	                seq_after(ESP_SKB_CB(skb)->ack_seq, esp_sk->send_seq)) {
		PR_DEBUG("esp socket %p received unexpected ack for seq %d, "
		         "expected (%d, %d]",
		         sk, ESP_SKB_CB(skb)->ack_seq, esp_sk->una_seq,
		         esp_sk->send_seq);
	}

	/* adjust minimum expected ACK sequence number */
	esp_sk->una_seq = ESP_SKB_CB(skb)->ack_seq;

	/* possibly we should move along the announce seq.
	 * to avoid a life-lock for really long transmissions
	 */
	if (!esp_waiting_for_first_ack(sk))
		esp_sk->announce_seq = esp_sk->una_seq - 2;

	return 1;
}

/*
 * Processes a received ACK and removes all data packets before the ACK from
 * the retransmission queue.
 * Returns zero if the ACK doesn't match any sent packet.
 */
static inline int esp_process_ack_late(struct sock *sk, struct sk_buff *skb)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sk_buff *skb1;

//   esp_push_pending_frames(sk);

	PR_DEBUG("Socket %p is processing an ACK, ack_seq: %d",
	         sk, ESP_SKB_CB(skb)->ack_seq);

	/* find all sent packets before this ACK and remove them from the
	 * retransmission queue */
	while ((skb1 = skb_peek(&sk->sk_write_queue)) != NULL) {
		if (!seq_before(ESP_SKB_CB(skb1)->pkt_seq,
		                ESP_SKB_CB(skb)->ack_seq))
			break;

		/* packet found
		 * Take the packet with the longest waiting time to
		 * update the socket's retransmission timeout */
		if (ESP_SKB_CB(skb1)->pkt_seq == esp_sk->una_seq &&
		                ESP_SKB_CB(skb1)->xmits == 1) {
			esp_update_rto(sk,
			               ESP_TIMESTAMP - ESP_SKB_CB(skb1)->sent_ts);
		}

		if (unlikely(sk->sk_send_head == skb1)) {
			/* the data was sent out by an RRQ, and now we got the ACK.
			 * -> move the send_head along.
			 */
			sk->sk_send_head = skb1->next;
			if (sk->sk_send_head == (struct sk_buff *)&sk->sk_write_queue)
				sk->sk_send_head = NULL;
		}

		/* dequeue and free the packet, reducing the amount of queued
		 * data
		 */
		__skb_unlink(skb1, &sk->sk_write_queue);

		sk_wmem_free_skb(sk, skb1);
	}

	/* restart retransmission timer if necessary */
	if (skb_queue_empty(&sk->sk_write_queue)) {
		//sk_stop_timer(sk, &esp_sk->rtx_timer);
	} else {
		//esp_rtx_timer_reset(sk, esp_sk->rto);
	}

	/* An ACK means the receiver is now responsible for
	 * requesting retransmissions, so we may stop our
	 * timer now.
	 */
	sk_stop_timer(sk, &esp_sk->rtx_timer);

	/* notify waiting processes about the new free space */
	sk_mem_reclaim(sk);
	sk->sk_write_space(sk);

	//read_lock(&sk->sk_callback_lock);

	//if (sk->sk_sleep && waitqueue_active(sk->sk_sleep)) {
	//printk("waking up a writer at %lu.\n", ESP_TIMESTAMP);
	//    wake_up_interruptible(sk->sk_sleep);
	//  }

	//  read_unlock(&sk->sk_callback_lock);

	PR_DEBUG("esp socket %p ack range (%d, %d] tx mem %d/%d "
	         "rtx queue len %d mem %d",
	         sk, esp_sk->una_seq, esp_sk->send_seq,
	         atomic_read(&sk->sk_wmem_alloc), sk->sk_sndbuf,
	         sk->sk_write_queue.qlen, sk->sk_wmem_queued);

	return 1;
}

/**
 * Does the early and the late phase of ACK processing together.
 */
static inline int esp_process_ack(struct sock* sk, struct sk_buff *skb)
{
	if (esp_process_ack_early(sk, skb)) {
		return esp_process_ack_late(sk, skb);
	}
	return 0;
}

/*
 * Processes a received RST
 * returns zero if the RST isn't acceptable
 */
static inline int esp_process_rst(struct sock *sk, struct sk_buff *skb,
                                  int err)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	switch (sk->sk_state) {
		/* ignore RST in the following socket states */
	case ESP_LISTEN:
		break;
		/* check the sequence number first in the following socket states */
	default:
		if (ESP_SKB_CB(skb)->pkt_seq != esp_sk->recv_seq) {
			return 0;
		}
		/* fall-through */
		/* reset the connection without sequence number check */
	case ESP_SYN_SENT:
		SK_ERR_REPORT(sk, err);
		esp_set_state(sk, ESP_CLOSE);
		__skb_queue_purge(&sk->sk_write_queue);
		if (!sock_flag(sk, SOCK_DEAD)) {
			sk->sk_state_change(sk);
		} else {
			esp_sk_destroy(sk);
		}
		break;
	}

	return 1;
}

/*
 * Processes a received FIN
 * must only be called in synchronized states,
 * returns zero if the FIN isn't acceptable
 */
static inline int esp_process_fin(struct sock *sk, struct sk_buff *skb)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* check if the FIN is acceptable */
	if (esp_sk->recv_seq != ESP_SKB_CB(skb)->pkt_seq)
		return 0;

	/* update the sequence number expected next to receive */
	printk("got a fin!\n");
	++esp_sk->recv_seq;

	/* ACK the FIN */
	esp_send_ack(sk);

	sk->sk_shutdown |= RCV_SHUTDOWN;
	sock_set_flag(sk, SOCK_DONE);

	switch (sk->sk_state) {
	case ESP_SYN_RECV:
	case ESP_ESTABLISHED: /* move to the CLOSE_WAIT state */
		esp_set_state(sk, ESP_CLOSE_WAIT);
		break;
	case ESP_FIN_WAIT1: /* simultaneous close */
		esp_set_state(sk, ESP_CLOSING);
		break;
	case ESP_FIN_WAIT2: /* peer also closes the connection */
		esp_set_state(sk, ESP_CLOSE);
		break;
	}

	/* clear the out-of-sequence queue if any data is left there */
	__skb_queue_purge(&esp_sk->oos_packets);
	sk_mem_reclaim(sk);

	if (!sock_flag(sk, SOCK_DEAD)) {
		sk->sk_state_change(sk);

		/* Do not send POLL_HUP for half duplex close. */
		if (sk->sk_shutdown == SHUTDOWN_MASK || sk->sk_state == ESP_CLOSE)
			sk_wake_async(sk, 1, POLL_HUP);
		else
			sk_wake_async(sk, 1, POLL_IN);
	} else {
		esp_set_state(sk, ESP_CLOSE);
		esp_sk_destroy(sk);
	}
	return 1;
}

/*
 * Enqueues a data packet onto the receiving queue
 * Returns
 *   0 if the packet wasn't enqueued and should be discarded
 *   1 if the packet was enqueued
 *  -1 if the packet was enqueued and oos-packets have been re-enqueued
 */
static inline int esp_enqueue_data(struct sock *sk, struct sk_buff *skb)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	unsigned int data_len;
	int rc = 0;

	/* update the maximum size of received socket buffers */
	if (unlikely(skb->truesize > esp_sk->skb_size)) {
		PR_DEBUG("esp socket %p updating max skb size %d -> %d",
		         sk, esp_sk->skb_size, skb->truesize);
		esp_sk->skb_size = skb->truesize;
	}

	/* check sequence number
	 * (they really should match so this is only to notice bugs) */
	if (unlikely(ESP_SKB_CB(skb)->pkt_seq != esp_sk->recv_seq)) {
		PR_ERR("esp socket %p tried to enqueue oos data to rx queue", sk);
		goto out;
	}

	/* zero byte data frames are not transmitted -
	 * so empty frames can be discarded (they are used with flags) */
	if (skb->len == 0)
		goto out;

	/* check the socket's receive memory */
	if (atomic_read(&sk->sk_rmem_alloc) + skb->truesize >=
	                (unsigned)sk->sk_rcvbuf) {
#if ESP_USE_OOS_QUEUE
		struct sk_buff *skb1;

		/* the out-of-sequence queue MUST be empty when this happens or
		 * else we got a bug producing a life lock ! */

		if (!skb_queue_empty(&esp_sk->oos_packets)) {
			skb1 = skb_peek(&esp_sk->oos_packets);
			PR_ERR("oos-q consumed all buffer space on esp socket %p "
			       "(RQ: %d mem %d/%d seq %d OQ: %d seq %d) "
			       "needed %d est %d left %d",
			       sk, sk->sk_receive_queue.qlen,
			       atomic_read(&sk->sk_rmem_alloc),
			       sk->sk_rcvbuf, esp_sk->recv_seq - 1,
			       esp_sk->oos_packets.qlen,
			       ESP_SKB_CB(skb1)->pkt_seq, skb->truesize,
			       esp_sk->skb_size,
			       (unsigned)sk->sk_rcvbuf - atomic_read(&sk->sk_rmem_alloc));
		}

		PR_DEBUG("esp socket %p out of receive memory "
		         "(RQ: %d mem %d/%d OQ: %d) - dropping packet # %d",
		         sk, sk->sk_receive_queue.qlen,
		         atomic_read(&sk->sk_rmem_alloc), sk->sk_rcvbuf,
		         esp_sk->oos_packets.qlen, ESP_SKB_CB(skb)->pkt_seq);
#endif

		++esp_sk->stats.pkt_rx_drop;

		/* uh-oh, we're in congestion situation. This could happen
		 * due to lots of unadvertised packets coming in...
		 */
		esp_enter_rcv_mem_pressure(sk);

		/* this was really acute and we can't handle any new data now,
		 * so we have to purge an ACK we could have on queue, too.
		 */
		esp_purge_ack(sk);

		PR_ERR("Uhh, I had to drop a packet!");

		goto out;
	}

	/* we will enqueue the packet */
	rc = 1;

	/* hand the socket buffer to the socket */
	skb_set_owner_r(skb, sk);
	__skb_queue_tail(&sk->sk_receive_queue, skb);

	++esp_sk->stats.pkt_rx;
	data_len = skb->len;

	/* increment next expected sequence number */
	++esp_sk->recv_seq;

	PR_DEBUG("enqueued received esp frame to socket %p "
	         "(RQ: %d mem %d/%d OQ: %d)",
	         sk, sk->sk_receive_queue.qlen, atomic_read(&sk->sk_rmem_alloc),
	         sk->sk_rcvbuf, esp_sk->oos_packets.qlen);

#if ESP_USE_OOS_QUEUE
	/* process out-of-seqence queue */
	for (skb1 = skb_peek(&esp_sk->oos_packets);
	                skb1 != NULL;
	                skb1 = skb_peek(&esp_sk->oos_packets)) {
		/* the queue is sorted - so stop if the packet doesn't
		 * bear the expected sequence number */
		if (ESP_SKB_CB(skb1)->pkt_seq != esp_sk->recv_seq)
			break;

		__skb_unlink(skb1, &esp_sk->oos_packets);

		if (skb1->len > 0) {
			/* re-enqueue out-of-sequence data packet */
			__skb_queue_tail(&sk->sk_receive_queue, skb1);
			data_len += skb1->len;

			rc = -1;

			PR_DEBUG("re-enqueued oos esp frame to socket %p "
			         "(RQ: %d mem %d/%d OQ: %d)",
			         sk, sk->sk_receive_queue.qlen,
			         atomic_read(&sk->sk_rmem_alloc), sk->sk_rcvbuf,
			         esp_sk->oos_packets.qlen);
		}

		if (ESP_HDR(skb1)->flags & ESP_FLAG_FIN) {
			esp_process_fin(sk, skb1);
			/* esp_process_fin() did purge the oos-queue */
			__kfree_skb(skb1);
			break;
		}

		/* increment next expected sequence number */
		++esp_sk->recv_seq;
	}
#endif

	/* report that there's data pending */
	sk->sk_data_ready(sk, data_len);

out:
	return rc;
}


#if ESP_USE_OOS_QUEUE
/*
 * Enqueues a data packet onto the receiving queue
 * Returns zero if the packet wasn't enqueued and should be discarded.
 */
static inline int esp_enqueue_oos(struct sock *sk, struct sk_buff *skb)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct esp_header *esp_hdr = ESP_HDR(skb);
	struct sk_buff *skb1;
	unsigned int pkt_hole, mem_hole;

	/* update the maximum size of received socket buffers */
	if (skb->truesize > esp_sk->skb_size) {
		PR_DEBUG("esp socket %p updating max skb size %d -> %d",
		         sk, esp_sk->skb_size, skb->truesize);
		esp_sk->skb_size = skb->truesize;
	}

	/* check sequence number
	 * (they really should match so this is only to notice bugs) */
	if (unlikely(!seq_after(ESP_SKB_CB(skb)->pkt_seq, esp_sk->recv_seq))) {
		PR_ERR("esp socket %p tried to enqueue invalid oos data", sk);
		goto discard;
	}

	/* discard empty non-FIN frames for they need not be kept */
	if (!skb->len && !(esp_hdr->flags & ESP_FLAG_FIN))
		goto discard;

	/* check the socket's receive memory - we MUST leave enough buffer
	 * space that the hole between the last in-sequence and the first
	 * out-of-sequence packet can be filled or we produce a dead lock */
	skb1 = skb_peek(&esp_sk->oos_packets);
	if (!skb1 || seq_before(ESP_SKB_CB(skb)->pkt_seq, ESP_SKB_CB(skb1)->pkt_seq))
		skb1 = skb;

	pkt_hole = ESP_SKB_CB(skb1)->pkt_seq - esp_sk->recv_seq + 1;
	mem_hole = pkt_hole * esp_sk->skb_size;

	PR_DEBUG("esp socket %p seq %d oos %d leaving %d bytes for "
	         "%d in-seq packets", sk, esp_sk->recv_seq - 1,
	         ESP_SKB_CB(skb1)->pkt_seq, mem_hole, pkt_hole);

	if (atomic_read(&sk->sk_rmem_alloc) + mem_hole + skb->truesize >=
	                (unsigned)sk->sk_rcvbuf) {
		++esp_sk->stats.pkt_rx_drop;
		PR_DEBUG("esp socket %p out of receive memory - "
		         "dropping oos packet %d",
		         sk, esp_sk->stats.pkt_rx_drop);
		goto discard;
	}

	/* most common case should be new data that goes to the end of the
	 * out-of-sequence queue */
	skb1 = skb_peek_tail(&esp_sk->oos_packets);

	/* is this the first out-of sequence packet ? */
	if (!skb1) {
		/* hand the socket buffer to the socket */
		skb_set_owner_r(skb, sk);
		__skb_queue_tail(&esp_sk->oos_packets, skb);
		goto queued;
	}

	/* find the place to insert the data */
	do {
		if (seq_after(ESP_SKB_CB(skb)->pkt_seq, ESP_SKB_CB(skb1)->pkt_seq))
			break;
	} while ((skb1 = skb1->prev) != (struct sk_buff*) & esp_sk->oos_packets);

	/* hand the socket buffer to the socket */
	skb_set_owner_r(skb, sk);
	__skb_insert(skb, skb1, skb1->next, &esp_sk->oos_packets);

queued:
	++esp_sk->stats.pkt_rx;
	++esp_sk->stats.pkt_rx_oos;

	PR_DEBUG("enqueued oos esp frame to socket %p (rq %d mem %d oq %d)",
	         sk, sk->sk_receive_queue.qlen, atomic_read(&sk->sk_rmem_alloc),
	         esp_sk->oos_packets.qlen);

	return 1;

discard:
	return 0;
}
#endif /* use OOS queue */

/* function prototypes */
int esp_rcv_packet(struct sock *sk, struct sk_buff *skb);
int sock_esp_recvmsg(struct kiocb *iocb, struct socket *sock,
                     struct msghdr *msg, size_t len, int flags);

#endif /* ENET_STREAM_INPUT_H_ */
