/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#ifndef PTABLE_H_
#define PTABLE_H_

#include <asm/page.h>  /* struct page */
#include <linux/swap.h>  /* mark_page_accessed() */
#include <asm/uaccess.h> /* get_user() */


/*
 * Determine the struct page address from the given memory management structure
 * (e.g. current->mm) and a correspondent virtual address.
 *
 * In more recent kernels (e.g. 2.6.13) a four level page table (pgd -> pud -> pmd -> pte)
 * is implemented, while older ones use only three levels (w/o pud)...
 *
 * WARNING: This function does NOT implement hugetlb pages !
 * see %KERNEL_SRC%/mm/memory.c :: follow_page(), which is not exported by the kernel...
 */
static inline struct page *lookup_page(struct mm_struct *mm, unsigned long address) {
	pgd_t *pgd;
	pmd_t *pmd;

#ifdef pud_offset
	/* 4-level page table needs page upper dir entry */
	pud_t *pud;
#endif

	pte_t *pte;
	unsigned long pfn;
	struct page *page = NULL;

	/* when a page is currently not in memory we do try paging just once */
	unsigned char paged = 0;

start:
	/* Get the page table lock */
	spin_lock(&mm->page_table_lock);

	/* Get page global directory (pgd) entry corresponding to address */
	pgd = pgd_offset(mm, address);
	if (pgd_none(*pgd) || unlikely(pgd_bad(*pgd))) {
		pgd_ERROR(*pgd);
		goto out_unlock;
	}

#ifdef pud_offset

	/* 4-level pagetable ... so process page upper dir first */

	/* Get page upper directory (pud) entry corresponding to the pgd entry and address */
	pud = pud_offset(pgd, address);
	if (pud_none(*pud) || unlikely(pud_bad(*pud))) {
		pud_ERROR(*pud);
		goto out_unlock;
	}
	/* Get page middle directory (pmd) entry corresponding to the pgd entry and address */
	pmd = pmd_offset(pud, address);
	if (pmd_none(*pmd) || unlikely(pmd_bad(*pmd))) {
		pmd_ERROR(*pmd);
		goto out_unlock;
	}

#else

	/* 3-level pagetable ... just go on with page middle dir */

	/* Get page middle directory (pmd) entry corresponding to the pgd entry and address */
	pmd = pmd_offset((pgd_t*)pgd, address);
	if (pmd_none(*pmd) || unlikely(pmd_bad(*pmd))) {
		pmd_ERROR(*pmd);
		goto out_unlock;
	}

#endif

	/* Get page table entry (pte) corresponding to the pmd entry and address */
	pte = pte_offset_map(pmd, address);
	if (!pte) {
		PR_ERR("invalid page table entry");
		goto out_unlock;
	}

	/* if the page is currently not in memory - try to page it in */
	if (unlikely(!pte_present(*pte))) {
		char _x;

		PR_DEBUG("page for address %#lx not present - trying to page it in", address);

		/* the process might sleep during paging -
		 * so unmap pte and release the page table lock */
		pte_unmap(pte);
		spin_unlock(&mm->page_table_lock);

		/* if paging has already been done and hasn't
		 * improved the situation - return no page */
		if (paged++)
			return NULL;

		/* get_user() activates paging on page fault */
		get_user(_x, (char __user *)address);

		goto start;
	}

	/* Determine the struct page address from the page table entry */
	if (pte_present(*pte) && pfn_valid(pfn = pte_pfn(*pte))) {
		page = pfn_to_page(pfn);
	}
	pte_unmap(pte);


	/* mark page accessed to prevent it from being swapped out soon */
	if (likely(page))
		mark_page_accessed(page);

out_unlock:
	/* release the page table lock */
	spin_unlock(&mm->page_table_lock);

	return page;
}


#endif /*PTABLE_H_*/
